"""Tests for MCP manager per-server lifecycle."""

from __future__ import annotations

import asyncio
import sys
from contextlib import AsyncExitStack
from typing import Any
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from anteroom.config import McpServerConfig
from anteroom.services.mcp_manager import McpManager


class TestMcpManagerInit:
    def test_configs_stored_as_dict(self) -> None:
        configs = [
            McpServerConfig(name="server-a", transport="stdio", command="echo"),
            McpServerConfig(name="server-b", transport="stdio", command="cat"),
        ]
        mgr = McpManager(configs)
        assert "server-a" in mgr._configs
        assert "server-b" in mgr._configs

    def test_empty_configs(self) -> None:
        mgr = McpManager([])
        assert mgr._configs == {}


class TestMcpManagerLifecycle:
    @pytest.fixture()
    def manager(self) -> McpManager:
        configs = [
            McpServerConfig(name="test-server", transport="stdio", command="echo"),
        ]
        return McpManager(configs)

    @pytest.mark.asyncio()
    async def test_startup_without_mcp_sdk(self, manager: McpManager) -> None:
        with patch.dict("sys.modules", {"mcp": None}):
            await manager.startup()
        assert manager.get_all_tools() == []

    @pytest.mark.asyncio()
    async def test_get_server_statuses_default(self, manager: McpManager) -> None:
        statuses = manager.get_server_statuses()
        assert "test-server" in statuses
        assert statuses["test-server"]["status"] == "disconnected"
        assert statuses["test-server"]["transport"] == "stdio"

    @pytest.mark.asyncio()
    async def test_disconnect_unknown_server_raises(self, manager: McpManager) -> None:
        with pytest.raises(ValueError, match="Unknown MCP server"):
            await manager.disconnect_server("nonexistent")

    @pytest.mark.asyncio()
    async def test_connect_unknown_server_raises(self, manager: McpManager) -> None:
        with pytest.raises(ValueError, match="Unknown MCP server"):
            await manager.connect_server("nonexistent")

    @pytest.mark.asyncio()
    async def test_get_openai_tools_empty(self, manager: McpManager) -> None:
        assert manager.get_openai_tools() is None

    @pytest.mark.asyncio()
    async def test_get_all_tools_empty(self, manager: McpManager) -> None:
        assert manager.get_all_tools() == []

    @pytest.mark.asyncio()
    async def test_shutdown_empty(self, manager: McpManager) -> None:
        await manager.shutdown()
        assert manager._sessions == {}
        assert manager._server_tools == {}

    def test_get_tool_server_name_unknown(self, manager: McpManager) -> None:
        assert manager.get_tool_server_name("nonexistent") == "unknown"


class TestMcpManagerToolMap:
    def test_rebuild_tool_map(self) -> None:
        mgr = McpManager([])
        mgr._server_tools = {
            "server-a": [
                {"name": "tool1", "server_name": "server-a", "description": "", "input_schema": {}},
                {"name": "tool2", "server_name": "server-a", "description": "", "input_schema": {}},
            ],
            "server-b": [
                {"name": "tool3", "server_name": "server-b", "description": "", "input_schema": {}},
            ],
        }
        mgr._rebuild_tool_map()
        assert mgr._tool_to_server == {"tool1": "server-a", "tool2": "server-a", "tool3": "server-b"}

    def test_get_all_tools_flattens(self) -> None:
        mgr = McpManager([])
        mgr._server_tools = {
            "server-a": [
                {"name": "tool1", "server_name": "server-a", "description": "d1", "input_schema": {}},
            ],
            "server-b": [
                {"name": "tool2", "server_name": "server-b", "description": "d2", "input_schema": {}},
            ],
        }
        tools = mgr.get_all_tools()
        assert len(tools) == 2
        names = {t["name"] for t in tools}
        assert names == {"tool1", "tool2"}

    def test_get_openai_tools_format(self) -> None:
        mgr = McpManager([])
        mgr._server_tools = {
            "server-a": [
                {
                    "name": "my_tool",
                    "server_name": "server-a",
                    "description": "Does stuff",
                    "input_schema": {"type": "object"},
                },
            ],
        }
        mgr._rebuild_tool_map()
        openai_tools = mgr.get_openai_tools()
        assert openai_tools is not None
        assert len(openai_tools) == 1
        assert openai_tools[0]["type"] == "function"
        assert openai_tools[0]["function"]["name"] == "my_tool"
        assert openai_tools[0]["function"]["description"] == "Does stuff"

    def test_disconnect_removes_tools(self) -> None:
        configs = [
            McpServerConfig(name="server-a", transport="stdio", command="echo"),
        ]
        mgr = McpManager(configs)
        mgr._server_tools = {
            "server-a": [
                {"name": "tool1", "server_name": "server-a", "description": "", "input_schema": {}},
            ],
        }
        mgr._tool_to_server = {"tool1": "server-a"}

        # Simulate disconnect's cleanup (without the async stack close)
        mgr._sessions.pop("server-a", None)
        mgr._server_tools.pop("server-a", None)
        mgr._rebuild_tool_map()

        assert mgr.get_all_tools() == []
        assert mgr._tool_to_server == {}


class TestMcpManagerCallTool:
    """Tests for call_tool error handling and argument passthrough."""

    @pytest.mark.asyncio()
    async def test_multiline_args_accepted(self) -> None:
        """MCP tool args with newlines (e.g. Jira descriptions) must not be rejected."""
        mgr = McpManager([])
        mgr._tool_to_server = {"update_issue": "jira"}

        mock_session = AsyncMock()
        mock_result = AsyncMock()
        mock_result.content = [AsyncMock(text="OK")]
        mock_session.call_tool = AsyncMock(return_value=mock_result)
        mgr._sessions = {"jira": mock_session}

        result = await mgr.call_tool(
            "update_issue",
            {
                "description": "Line 1\nLine 2\nLine 3\r\nLine 4",
                "summary": "Test with special chars: $100 & (parens) {braces}",
            },
        )
        assert result["content"] == "OK"
        mock_session.call_tool.assert_awaited_once()

    @pytest.mark.asyncio()
    async def test_tool_not_found_lists_connected_servers(self) -> None:
        """Error for missing tool should list connected servers."""
        mgr = McpManager([])
        mgr._sessions = {"jira": AsyncMock(), "slack": AsyncMock()}

        with pytest.raises(ValueError, match="MCP tool 'nonexistent' not available") as exc_info:
            await mgr.call_tool("nonexistent", {})

        msg = str(exc_info.value)
        assert "jira" in msg
        assert "slack" in msg

    @pytest.mark.asyncio()
    async def test_tool_not_found_no_servers(self) -> None:
        """Error for missing tool with no connected servers shows (none)."""
        mgr = McpManager([])
        mgr._sessions = {}

        with pytest.raises(ValueError, match=r"\(none\)"):
            await mgr.call_tool("anything", {})

    @pytest.mark.asyncio()
    async def test_server_error_includes_context(self) -> None:
        """Server-side errors should include server name and tool name."""
        mgr = McpManager([])
        mgr._tool_to_server = {"broken_tool": "my-server"}

        mock_session = AsyncMock()
        mock_session.call_tool = AsyncMock(side_effect=RuntimeError("connection reset"))
        mgr._sessions = {"my-server": mock_session}

        with pytest.raises(ValueError, match="MCP server 'my-server'") as exc_info:
            await mgr.call_tool("broken_tool", {"arg": "value"})

        msg = str(exc_info.value)
        assert "broken_tool" in msg
        assert "connection reset" not in msg  # raw exception not exposed to caller

    @pytest.mark.asyncio()
    async def test_server_error_chains_original(self) -> None:
        """Server-side errors should chain the original exception."""
        mgr = McpManager([])
        mgr._tool_to_server = {"my_tool": "srv"}

        original = RuntimeError("original error")
        mock_session = AsyncMock()
        mock_session.call_tool = AsyncMock(side_effect=original)
        mgr._sessions = {"srv": mock_session}

        with pytest.raises(ValueError) as exc_info:
            await mgr.call_tool("my_tool", {})

        assert exc_info.value.__cause__ is original


class TestMcpManagerFailedConnection:
    """Tests for MCP connection failures — stack must be cleaned up."""

    @pytest.mark.asyncio()
    async def test_connect_one_exception_closes_stack(self) -> None:
        """When _connect_one raises, the AsyncExitStack must be closed."""
        config = McpServerConfig(name="bad-server", transport="stdio", command="echo")
        mgr = McpManager([config])

        mock_stack = AsyncMock(spec=AsyncExitStack)
        mock_stack.enter_async_context = AsyncMock(side_effect=ConnectionError("server crashed"))

        with (
            patch("anteroom.services.mcp_manager.AsyncExitStack", return_value=mock_stack),
            patch("anteroom.services.mcp_manager.shutil.which", return_value="/usr/bin/echo"),
        ):
            await mgr._connect_one(config)

        # Stack must have been closed to prevent leaked task groups
        mock_stack.aclose.assert_awaited_once()
        # Server should be marked as error
        assert mgr._server_status["bad-server"]["status"] == "error"
        assert "server crashed" in mgr._server_status["bad-server"]["error_message"]
        # No session or tools left behind
        assert "bad-server" not in mgr._sessions
        assert "bad-server" not in mgr._exit_stacks

    @pytest.mark.asyncio()
    @pytest.mark.skipif(sys.version_info < (3, 11), reason="ExceptionGroup requires 3.11+")
    async def test_connect_one_exception_group_closes_stack(self) -> None:
        """ExceptionGroup from a TaskGroup must also close the stack."""
        config = McpServerConfig(name="bad-server", transport="stdio", command="echo")
        mgr = McpManager([config])

        exc_group = ExceptionGroup("task group failed", [RuntimeError("subtask died")])  # noqa: F821
        mock_stack = AsyncMock(spec=AsyncExitStack)
        mock_stack.enter_async_context = AsyncMock(side_effect=exc_group)

        with (
            patch("anteroom.services.mcp_manager.AsyncExitStack", return_value=mock_stack),
            patch("anteroom.services.mcp_manager.shutil.which", return_value="/usr/bin/echo"),
        ):
            await mgr._connect_one(config)

        mock_stack.aclose.assert_awaited_once()
        assert mgr._server_status["bad-server"]["status"] == "error"

    @pytest.mark.asyncio()
    async def test_startup_continues_after_one_server_fails(self) -> None:
        """If one server fails to connect, the others should still connect."""
        configs = [
            McpServerConfig(name="good-server", transport="stdio", command="echo"),
            McpServerConfig(name="bad-server", transport="stdio", command="echo"),
        ]
        mgr = McpManager(configs)

        call_count = 0

        async def fake_connect_one(config: McpServerConfig, status_callback: Any = None) -> None:
            nonlocal call_count
            call_count += 1
            if config.name == "bad-server":
                mgr._server_status[config.name] = {
                    "status": "error",
                    "tool_count": 0,
                    "error_message": "connection refused",
                }
            else:
                mgr._server_status[config.name] = {
                    "status": "connected",
                    "tool_count": 3,
                }

        with patch.object(mgr, "_connect_one", side_effect=fake_connect_one):
            with patch.dict("sys.modules", {"mcp": MagicMock(), "mcp.client.stdio": MagicMock()}):
                await mgr.startup()

        assert call_count == 2
        assert mgr._server_status["good-server"]["status"] == "connected"
        assert mgr._server_status["bad-server"]["status"] == "error"

    @pytest.mark.asyncio()
    async def test_connect_one_stack_aclose_failure_still_sets_error(self) -> None:
        """Even if stack.aclose() itself fails, server status should be set."""
        config = McpServerConfig(name="messy-server", transport="stdio", command="echo")
        mgr = McpManager([config])

        mock_stack = AsyncMock(spec=AsyncExitStack)
        mock_stack.enter_async_context = AsyncMock(side_effect=OSError("broken pipe"))
        mock_stack.aclose = AsyncMock(side_effect=RuntimeError("cleanup exploded"))

        with (
            patch("anteroom.services.mcp_manager.AsyncExitStack", return_value=mock_stack),
            patch("anteroom.services.mcp_manager.shutil.which", return_value="/usr/bin/echo"),
        ):
            await mgr._connect_one(config)

        # Despite aclose failing, status should still be set
        assert mgr._server_status["messy-server"]["status"] == "error"
        assert "broken pipe" in mgr._server_status["messy-server"]["error_message"]

    @pytest.mark.asyncio()
    async def test_connect_one_invalid_transport_does_not_leak_stack(self) -> None:
        """Invalid transport config should return early without leaking."""
        config = McpServerConfig(name="bad-transport", transport="invalid", command="")
        mgr = McpManager([config])

        await mgr._connect_one(config)

        assert mgr._server_status["bad-transport"]["status"] == "error"
        assert "Invalid transport" in mgr._server_status["bad-transport"]["error_message"]
        assert "bad-transport" not in mgr._exit_stacks

    @pytest.mark.asyncio()
    async def test_failed_server_has_no_tools(self) -> None:
        """A failed server should contribute zero tools."""
        configs = [
            McpServerConfig(name="dead-server", transport="stdio", command="echo"),
        ]
        mgr = McpManager(configs)

        mock_stack = AsyncMock(spec=AsyncExitStack)
        mock_stack.enter_async_context = AsyncMock(side_effect=ConnectionRefusedError("refused"))

        with (
            patch("anteroom.services.mcp_manager.AsyncExitStack", return_value=mock_stack),
            patch("anteroom.services.mcp_manager.shutil.which", return_value="/usr/bin/echo"),
        ):
            await mgr._connect_one(configs[0])

        assert mgr.get_all_tools() == []
        assert mgr.get_openai_tools() is None
        assert mgr.get_tool_server_name("any_tool") == "unknown"

    @pytest.mark.asyncio()
    async def test_error_message_includes_exception_type(self) -> None:
        """Error message should include the exception class name for diagnostics."""
        config = McpServerConfig(name="typed-error", transport="stdio", command="echo")
        mgr = McpManager([config])

        mock_stack = AsyncMock(spec=AsyncExitStack)
        mock_stack.enter_async_context = AsyncMock(side_effect=TimeoutError("timed out after 30s"))

        with (
            patch("anteroom.services.mcp_manager.AsyncExitStack", return_value=mock_stack),
            patch("anteroom.services.mcp_manager.shutil.which", return_value="/usr/bin/echo"),
        ):
            await mgr._connect_one(config)

        err = mgr._server_status["typed-error"]["error_message"]
        assert "TimeoutError" in err
        assert "timed out after 30s" in err

    @pytest.mark.asyncio()
    async def test_command_not_found_shows_path(self) -> None:
        """FileNotFoundError should mention PATH when command is missing."""
        config = McpServerConfig(name="missing-cmd", transport="stdio", command="nonexistent-mcp-server")
        mgr = McpManager([config])

        with patch("anteroom.services.mcp_manager.shutil.which", return_value=None):
            await mgr._connect_one(config)

        err = mgr._server_status["missing-cmd"]["error_message"]
        assert "not found on PATH" in err
        assert "nonexistent-mcp-server" in err

    def test_describe_config_stdio(self) -> None:
        """_describe_config should include command and args."""
        config = McpServerConfig(name="jira", transport="stdio", command="npx", args=["-y", "jira-mcp-server"])
        mgr = McpManager([config])
        desc = mgr._describe_config(config)
        assert "jira" in desc
        assert "stdio" in desc
        assert "npx -y jira-mcp-server" in desc

    def test_describe_config_sse(self) -> None:
        """_describe_config should include URL for SSE."""
        config = McpServerConfig(name="remote", transport="sse", url="https://mcp.example.com/sse")
        mgr = McpManager([config])
        desc = mgr._describe_config(config)
        assert "sse" in desc
        assert "https://mcp.example.com/sse" in desc

    def test_describe_config_with_env(self) -> None:
        """_describe_config should list env key names (not values)."""
        config = McpServerConfig(
            name="secret", transport="stdio", command="cmd", env={"API_KEY": "sk-123", "TOKEN": "abc"}
        )
        mgr = McpManager([config])
        desc = mgr._describe_config(config)
        assert "API_KEY" in desc
        assert "TOKEN" in desc
        assert "sk-123" not in desc  # values must not leak

    @pytest.mark.asyncio()
    async def test_connection_timeout(self) -> None:
        """Server that hangs should be killed after timeout."""
        config = McpServerConfig(name="slow-server", transport="stdio", command="echo", timeout=0.1)
        mgr = McpManager([config])

        async def hang_forever(_config: McpServerConfig) -> None:
            await asyncio.sleep(999)

        with patch.object(mgr, "_do_connect", side_effect=hang_forever):
            await mgr._connect_one(config)

        assert mgr._server_status["slow-server"]["status"] == "error"
        assert "timed out" in mgr._server_status["slow-server"]["error_message"].lower()

    @pytest.mark.asyncio()
    async def test_custom_timeout_from_config(self) -> None:
        """Timeout value should come from config."""
        config = McpServerConfig(name="custom", transport="stdio", command="echo", timeout=60.0)
        mgr = McpManager([config])
        assert mgr._configs["custom"].timeout == 60.0

    def test_default_timeout(self) -> None:
        """Default timeout should be 30s."""
        config = McpServerConfig(name="default", transport="stdio", command="echo")
        assert config.timeout == 30.0

    def test_describe_config_includes_timeout(self) -> None:
        """_describe_config should show timeout."""
        config = McpServerConfig(name="t", transport="stdio", command="echo", timeout=15.0)
        mgr = McpManager([config])
        desc = mgr._describe_config(config)
        assert "timeout=15.0s" in desc


class TestMcpManagerErrorClassification:
    """Tests for exc_info suppression on McpError vs full traceback for unexpected errors."""

    @pytest.mark.asyncio()
    async def test_mcp_error_suppresses_traceback(self) -> None:
        """McpError should be logged without exc_info so no raw traceback appears."""
        try:
            from mcp import McpError
            from mcp.types import ErrorData
        except ImportError:
            pytest.skip("mcp SDK not installed")

        config = McpServerConfig(name="mcp-err-server", transport="stdio", command="echo")
        mgr = McpManager([config])
        logged_kwargs: list[dict] = []

        import logging

        mcp_error = McpError(ErrorData(code=-32000, message="server rejected handshake"))

        with (
            patch("anteroom.services.mcp_manager.shutil.which", return_value="/usr/bin/echo"),
            patch("anteroom.services.mcp_manager.AsyncExitStack") as mock_stack_cls,
            patch.object(
                logging.getLogger("anteroom.services.mcp_manager"),
                "warning",
                side_effect=lambda *a, **kw: logged_kwargs.append(kw),
            ),
        ):
            mock_stack = AsyncMock()
            mock_stack_cls.return_value = mock_stack
            mock_stack.enter_async_context.side_effect = mcp_error
            mock_stack.aclose = AsyncMock()
            await mgr._do_connect(config)

        error_calls = [kw for kw in logged_kwargs if "exc_info" in kw]
        assert error_calls, "Expected at least one logger.warning call with exc_info kwarg"
        assert error_calls[-1]["exc_info"] is False, "McpError should suppress traceback (exc_info=False)"

    @pytest.mark.asyncio()
    async def test_unexpected_error_keeps_traceback(self) -> None:
        """Non-McpError exceptions should be logged with exc_info=True."""
        config = McpServerConfig(name="unexpected-server", transport="stdio", command="echo")
        mgr = McpManager([config])
        logged_kwargs: list[dict] = []

        import logging

        with (
            patch("anteroom.services.mcp_manager.shutil.which", return_value="/usr/bin/echo"),
            patch("anteroom.services.mcp_manager.AsyncExitStack") as mock_stack_cls,
            patch.object(
                logging.getLogger("anteroom.services.mcp_manager"),
                "warning",
                side_effect=lambda *a, **kw: logged_kwargs.append(kw),
            ),
        ):
            mock_stack = AsyncMock()
            mock_stack_cls.return_value = mock_stack
            mock_stack.enter_async_context.side_effect = RuntimeError("segfault in subprocess")
            mock_stack.aclose = AsyncMock()
            await mgr._do_connect(config)

        error_calls = [kw for kw in logged_kwargs if "exc_info" in kw]
        assert error_calls, "Expected at least one logger.warning call with exc_info kwarg"
        assert error_calls[-1]["exc_info"] is True, "Unexpected errors should include traceback (exc_info=True)"

    @pytest.mark.asyncio()
    async def test_cancelled_error_reraised_after_stack_cleanup(self) -> None:
        """CancelledError must propagate out of _do_connect so wait_for works."""
        config = McpServerConfig(name="cancelled-server", transport="stdio", command="echo")
        mgr = McpManager([config])
        stack_closed = False

        with (
            patch("anteroom.services.mcp_manager.shutil.which", return_value="/usr/bin/echo"),
            patch("anteroom.services.mcp_manager.AsyncExitStack") as mock_stack_cls,
        ):
            mock_stack = AsyncMock()

            async def close_and_track():
                nonlocal stack_closed
                stack_closed = True

            mock_stack.aclose = close_and_track
            mock_stack_cls.return_value = mock_stack
            mock_stack.enter_async_context.side_effect = asyncio.CancelledError()

            with pytest.raises(asyncio.CancelledError):
                await mgr._do_connect(config)

        assert stack_closed, "Stack should be closed even when CancelledError is re-raised"
        # Server status should NOT be set to error — cancellation is not a connection failure
        assert "cancelled-server" not in mgr._server_status


class TestMcpManagerShutdown:
    """Tests for shutdown handling of BaseException and ExceptionGroup."""

    @pytest.mark.asyncio()
    async def test_shutdown_closes_active_stacks(self) -> None:
        """Shutdown should close all active exit stacks."""
        mgr = McpManager([])
        stack_a = AsyncMock(spec=AsyncExitStack)
        stack_b = AsyncMock(spec=AsyncExitStack)
        mgr._exit_stacks = {"server-a": stack_a, "server-b": stack_b}
        mgr._sessions = {"server-a": AsyncMock(), "server-b": AsyncMock()}

        await mgr.shutdown()

        stack_a.aclose.assert_awaited_once()
        stack_b.aclose.assert_awaited_once()
        assert mgr._exit_stacks == {}
        assert mgr._sessions == {}

    @pytest.mark.asyncio()
    async def test_shutdown_handles_base_exception_from_stack(self) -> None:
        """Shutdown must catch BaseException (including ExceptionGroup) from stack.aclose()."""
        mgr = McpManager([])
        stack = AsyncMock(spec=AsyncExitStack)
        stack.aclose = AsyncMock(side_effect=BaseException("task group teardown"))
        mgr._exit_stacks = {"server": stack}
        mgr._sessions = {"server": AsyncMock()}

        await mgr.shutdown()

        assert mgr._exit_stacks == {}
        assert mgr._sessions == {}

    @pytest.mark.asyncio()
    @pytest.mark.skipif(sys.version_info < (3, 11), reason="ExceptionGroup requires 3.11+")
    async def test_shutdown_handles_exception_group(self) -> None:
        """Shutdown must survive ExceptionGroup from TaskGroup teardown."""
        mgr = McpManager([])
        exc_group = ExceptionGroup("task group failed", [RuntimeError("subtask died")])  # noqa: F821
        stack = AsyncMock(spec=AsyncExitStack)
        stack.aclose = AsyncMock(side_effect=exc_group)
        mgr._exit_stacks = {"server": stack}
        mgr._sessions = {"server": AsyncMock()}

        await mgr.shutdown()

        assert mgr._exit_stacks == {}
        assert mgr._sessions == {}

    @pytest.mark.asyncio()
    async def test_shutdown_timeout_prevents_hang(self) -> None:
        """Shutdown should complete within its internal 5s timeout, not hang forever."""
        mgr = McpManager([])
        entered_hang = asyncio.Event()

        async def hang_forever() -> None:
            entered_hang.set()
            await asyncio.sleep(999)

        stack = AsyncMock(spec=AsyncExitStack)
        stack.aclose = hang_forever
        mgr._exit_stacks = {"stuck-server": stack}
        mgr._sessions = {"stuck-server": AsyncMock()}

        import time

        start = time.monotonic()
        await mgr.shutdown()
        elapsed = time.monotonic() - start

        assert mgr._exit_stacks == {}
        assert entered_hang.is_set(), "hang_forever should have been entered"
        assert elapsed < 8.0, f"Shutdown took {elapsed:.1f}s — internal timeout not working"

    @pytest.mark.asyncio()
    async def test_disconnect_handles_base_exception_from_stack(self) -> None:
        """disconnect_server must catch BaseException from stack.aclose()."""
        config = McpServerConfig(name="server", transport="stdio", command="echo")
        mgr = McpManager([config])
        stack = AsyncMock(spec=AsyncExitStack)
        stack.aclose = AsyncMock(side_effect=BaseException("task group teardown"))
        mgr._exit_stacks = {"server": stack}
        mgr._sessions = {"server": AsyncMock()}

        await mgr.disconnect_server("server")

        assert mgr._server_status["server"]["status"] == "disconnected"
        assert "server" not in mgr._exit_stacks

    @pytest.mark.asyncio()
    @pytest.mark.skipif(sys.version_info < (3, 11), reason="ExceptionGroup requires 3.11+")
    async def test_disconnect_handles_exception_group(self) -> None:
        """disconnect_server must survive ExceptionGroup from TaskGroup teardown."""
        config = McpServerConfig(name="server", transport="stdio", command="echo")
        mgr = McpManager([config])
        exc_group = ExceptionGroup("task group failed", [RuntimeError("subtask died")])  # noqa: F821
        stack = AsyncMock(spec=AsyncExitStack)
        stack.aclose = AsyncMock(side_effect=exc_group)
        mgr._exit_stacks = {"server": stack}
        mgr._sessions = {"server": AsyncMock()}

        await mgr.disconnect_server("server")

        assert mgr._server_status["server"]["status"] == "disconnected"
        assert "server" not in mgr._exit_stacks

    @pytest.mark.asyncio()
    async def test_shutdown_logs_errors_at_debug_not_warning(self, caplog: pytest.LogCaptureFixture) -> None:
        """Shutdown errors should be logged at DEBUG, not WARNING, to avoid noisy tracebacks on Ctrl+C."""
        import logging

        mgr = McpManager([])
        stack = AsyncMock(spec=AsyncExitStack)
        stack.aclose = AsyncMock(side_effect=RuntimeError("Attempted to exit a cancel scope"))
        mgr._exit_stacks = {"clickup": stack}
        mgr._sessions = {"clickup": AsyncMock()}

        with caplog.at_level(logging.DEBUG, logger="anteroom.services.mcp_manager"):
            await mgr.shutdown()

        shutdown_records = [r for r in caplog.records if "clickup" in r.message and "shutdown" in r.message]
        assert shutdown_records, "Expected a log record mentioning 'clickup' and 'shutdown'"
        assert all(r.levelno == logging.DEBUG for r in shutdown_records), (
            f"Expected DEBUG level, got: {[r.levelname for r in shutdown_records]}"
        )
        warning_records = [r for r in caplog.records if r.levelno >= logging.WARNING and "clickup" in r.message]
        assert not warning_records, f"Unexpected WARNING+ records: {warning_records}"

    @pytest.mark.asyncio()
    async def test_shutdown_suppresses_cancelled_error(self) -> None:
        """CancelledError during shutdown (common on Ctrl+C) should not propagate."""
        mgr = McpManager([])
        stack = AsyncMock(spec=AsyncExitStack)
        stack.aclose = AsyncMock(side_effect=asyncio.CancelledError())
        mgr._exit_stacks = {"time": stack}
        mgr._sessions = {"time": AsyncMock()}

        await mgr.shutdown()

        stack.aclose.assert_awaited_once()
        assert mgr._exit_stacks == {}

    @pytest.mark.asyncio()
    @pytest.mark.skipif(
        sys.version_info < (3, 12),
        reason="asyncio.wait_for propagates KeyboardInterrupt at the event loop level on <3.12",
    )
    async def test_shutdown_reraises_keyboard_interrupt_after_cleanup(self) -> None:
        """KeyboardInterrupt during shutdown should be re-raised after all stacks are cleaned."""
        mgr = McpManager([])

        # Use a real async function rather than AsyncMock side_effect for
        # KeyboardInterrupt — AsyncMock's side_effect handling for BaseException
        # subclasses leaks through asyncio.wait_for on Python <3.12.
        stack_a_closed = False
        stack_b_closed = False

        async def _aclose_raises_ki() -> None:
            nonlocal stack_a_closed
            stack_a_closed = True
            raise KeyboardInterrupt()

        async def _aclose_ok() -> None:
            nonlocal stack_b_closed
            stack_b_closed = True

        stack_a = AsyncMock(spec=AsyncExitStack)
        stack_a.aclose = _aclose_raises_ki
        stack_b = AsyncMock(spec=AsyncExitStack)
        stack_b.aclose = _aclose_ok
        mgr._exit_stacks = {"server-a": stack_a, "server-b": stack_b}
        mgr._sessions = {"server-a": AsyncMock(), "server-b": AsyncMock()}

        with pytest.raises(KeyboardInterrupt):
            await mgr.shutdown()

        # Both stacks must have been closed before re-raising
        assert stack_a_closed
        assert stack_b_closed
        assert mgr._exit_stacks == {}
        assert mgr._sessions == {}


def _make_tool(name: str, server: str) -> dict[str, Any]:
    return {"name": name, "server_name": server, "description": f"desc-{name}", "input_schema": {}}


class TestMcpToolFiltering:
    """Tests for per-server tool include/exclude filtering."""

    def _make_manager(self, *configs: McpServerConfig) -> McpManager:
        mgr = McpManager(list(configs))
        return mgr

    def test_no_filter_returns_all_tools(self) -> None:
        cfg = McpServerConfig(name="jira", transport="stdio", command="echo")
        mgr = self._make_manager(cfg)
        mgr._server_tools = {"jira": [_make_tool("search", "jira"), _make_tool("create", "jira")]}
        mgr._rebuild_tool_map()
        assert len(mgr.get_all_tools()) == 2

    def test_include_filters_to_allowlist(self) -> None:
        cfg = McpServerConfig(name="jira", transport="stdio", command="echo", tools_include=["search", "get_issue"])
        mgr = self._make_manager(cfg)
        mgr._server_tools = {
            "jira": [_make_tool("search", "jira"), _make_tool("create", "jira"), _make_tool("get_issue", "jira")],
        }
        mgr._rebuild_tool_map()
        names = {t["name"] for t in mgr.get_all_tools()}
        assert names == {"search", "get_issue"}

    def test_exclude_filters_out_blocklist(self) -> None:
        cfg = McpServerConfig(name="jira", transport="stdio", command="echo", tools_exclude=["delete", "bulk_delete"])
        mgr = self._make_manager(cfg)
        mgr._server_tools = {
            "jira": [_make_tool("search", "jira"), _make_tool("delete", "jira"), _make_tool("bulk_delete", "jira")],
        }
        mgr._rebuild_tool_map()
        names = {t["name"] for t in mgr.get_all_tools()}
        assert names == {"search"}

    def test_include_glob_pattern(self) -> None:
        cfg = McpServerConfig(name="jira", transport="stdio", command="echo", tools_include=["get_*", "search_*"])
        mgr = self._make_manager(cfg)
        mgr._server_tools = {
            "jira": [
                _make_tool("get_issue", "jira"),
                _make_tool("get_comment", "jira"),
                _make_tool("search_issues", "jira"),
                _make_tool("create_issue", "jira"),
                _make_tool("delete_issue", "jira"),
            ],
        }
        mgr._rebuild_tool_map()
        names = {t["name"] for t in mgr.get_all_tools()}
        assert names == {"get_issue", "get_comment", "search_issues"}

    def test_exclude_glob_pattern(self) -> None:
        cfg = McpServerConfig(name="jira", transport="stdio", command="echo", tools_exclude=["bulk_*", "admin_*"])
        mgr = self._make_manager(cfg)
        mgr._server_tools = {
            "jira": [
                _make_tool("search", "jira"),
                _make_tool("bulk_create", "jira"),
                _make_tool("bulk_delete", "jira"),
                _make_tool("admin_settings", "jira"),
            ],
        }
        mgr._rebuild_tool_map()
        names = {t["name"] for t in mgr.get_all_tools()}
        assert names == {"search"}

    def test_filtered_tool_not_in_tool_map(self) -> None:
        cfg = McpServerConfig(name="jira", transport="stdio", command="echo", tools_include=["search"])
        mgr = self._make_manager(cfg)
        mgr._server_tools = {"jira": [_make_tool("search", "jira"), _make_tool("delete", "jira")]}
        mgr._rebuild_tool_map()
        assert mgr.get_tool_server_name("search") == "jira"
        assert mgr.get_tool_server_name("delete") == "unknown"

    def test_get_openai_tools_respects_filter(self) -> None:
        cfg = McpServerConfig(name="jira", transport="stdio", command="echo", tools_include=["search"])
        mgr = self._make_manager(cfg)
        mgr._server_tools = {"jira": [_make_tool("search", "jira"), _make_tool("delete", "jira")]}
        mgr._rebuild_tool_map()
        openai_tools = mgr.get_openai_tools()
        assert openai_tools is not None
        assert len(openai_tools) == 1
        assert openai_tools[0]["function"]["name"] == "search"

    def test_multi_server_independent_filters(self) -> None:
        cfg_a = McpServerConfig(name="jira", transport="stdio", command="echo", tools_include=["search"])
        cfg_b = McpServerConfig(name="confluence", transport="stdio", command="echo", tools_exclude=["delete_*"])
        mgr = self._make_manager(cfg_a, cfg_b)
        mgr._server_tools = {
            "jira": [_make_tool("search", "jira"), _make_tool("create", "jira")],
            "confluence": [_make_tool("read_page", "confluence"), _make_tool("delete_page", "confluence")],
        }
        mgr._rebuild_tool_map()
        names = {t["name"] for t in mgr.get_all_tools()}
        assert names == {"search", "read_page"}

    @pytest.mark.asyncio()
    async def test_call_tool_blocked_for_filtered_tool(self) -> None:
        cfg = McpServerConfig(name="jira", transport="stdio", command="echo", tools_include=["search"])
        mgr = self._make_manager(cfg)
        mgr._server_tools = {"jira": [_make_tool("search", "jira"), _make_tool("delete", "jira")]}
        mgr._sessions = {"jira": AsyncMock()}
        mgr._rebuild_tool_map()
        with pytest.raises(ValueError, match="filtered out"):
            await mgr.call_tool("delete", {})

    def test_empty_include_means_no_filter(self) -> None:
        cfg = McpServerConfig(name="jira", transport="stdio", command="echo", tools_include=[])
        mgr = self._make_manager(cfg)
        mgr._server_tools = {"jira": [_make_tool("search", "jira"), _make_tool("create", "jira")]}
        mgr._rebuild_tool_map()
        assert len(mgr.get_all_tools()) == 2

    def test_empty_exclude_means_no_filter(self) -> None:
        cfg = McpServerConfig(name="jira", transport="stdio", command="echo", tools_exclude=[])
        mgr = self._make_manager(cfg)
        mgr._server_tools = {"jira": [_make_tool("search", "jira"), _make_tool("create", "jira")]}
        mgr._rebuild_tool_map()
        assert len(mgr.get_all_tools()) == 2

    @pytest.mark.asyncio()
    async def test_tool_warning_threshold_fires(self, caplog: pytest.LogCaptureFixture) -> None:
        import logging

        cfg = McpServerConfig(name="jira", transport="stdio", command="echo")
        mgr = McpManager([cfg], tool_warning_threshold=3)
        mgr._server_tools = {"jira": [_make_tool(f"tool{i}", "jira") for i in range(5)]}
        mgr._rebuild_tool_map()
        mgr._server_status = {"jira": {"status": "connected", "tool_count": 5}}

        async def noop(_config: McpServerConfig) -> None:
            pass

        with (
            caplog.at_level(logging.WARNING, logger="anteroom.services.mcp_manager"),
            patch.object(mgr, "_connect_one", side_effect=noop),
        ):
            await mgr.startup()

        warning_records = [r for r in caplog.records if "exceeds threshold" in r.message]
        assert warning_records, "Expected tool threshold warning"
        assert "5" in warning_records[0].message
        assert "3" in warning_records[0].message

    @pytest.mark.asyncio()
    async def test_tool_warning_threshold_disabled(self, caplog: pytest.LogCaptureFixture) -> None:
        import logging

        cfg = McpServerConfig(name="jira", transport="stdio", command="echo")
        mgr = McpManager([cfg], tool_warning_threshold=0)
        mgr._server_tools = {"jira": [_make_tool(f"tool{i}", "jira") for i in range(50)]}
        mgr._rebuild_tool_map()
        mgr._server_status = {"jira": {"status": "connected", "tool_count": 50}}

        async def noop(_config: McpServerConfig) -> None:
            pass

        with (
            caplog.at_level(logging.WARNING, logger="anteroom.services.mcp_manager"),
            patch.object(mgr, "_connect_one", side_effect=noop),
        ):
            await mgr.startup()

        warning_records = [r for r in caplog.records if "exceeds threshold" in r.message]
        assert not warning_records, "Should not warn when threshold is 0 (disabled)"

    @pytest.mark.asyncio()
    async def test_tool_warning_threshold_not_exceeded(self, caplog: pytest.LogCaptureFixture) -> None:
        import logging

        cfg = McpServerConfig(name="jira", transport="stdio", command="echo")
        mgr = McpManager([cfg], tool_warning_threshold=10)
        mgr._server_tools = {"jira": [_make_tool(f"tool{i}", "jira") for i in range(5)]}
        mgr._rebuild_tool_map()
        mgr._server_status = {"jira": {"status": "connected", "tool_count": 5}}

        async def noop(_config: McpServerConfig) -> None:
            pass

        with (
            caplog.at_level(logging.WARNING, logger="anteroom.services.mcp_manager"),
            patch.object(mgr, "_connect_one", side_effect=noop),
        ):
            await mgr.startup()

        warning_records = [r for r in caplog.records if "exceeds threshold" in r.message]
        assert not warning_records, "Should not warn when under threshold"

    def test_include_no_matches_returns_empty(self) -> None:
        cfg = McpServerConfig(name="jira", transport="stdio", command="echo", tools_include=["nonexistent_*"])
        mgr = self._make_manager(cfg)
        mgr._server_tools = {"jira": [_make_tool("search", "jira"), _make_tool("create", "jira")]}
        mgr._rebuild_tool_map()
        assert mgr.get_all_tools() == []
        assert mgr.get_openai_tools() is None


# Additional tests for coverage of missed lines (#689)


class TestValidateSseUrl:
    """Tests for _validate_sse_url covering lines 37-59."""

    def test_valid_https_url_passes(self) -> None:
        from anteroom.services.mcp_manager import _validate_sse_url

        # Should not raise — external reachable hostname
        # We need a URL that won't be SSRF-blocked; use a known public IP
        # but since DNS is live, use an IP directly instead
        _validate_sse_url("https://93.184.216.34/sse")  # example.com IP

    def test_invalid_scheme_raises(self) -> None:
        from anteroom.services.mcp_manager import _validate_sse_url

        with pytest.raises(ValueError, match="Unsupported URL scheme"):
            _validate_sse_url("ftp://example.com/sse")

    def test_localhost_hostname_blocked(self) -> None:
        from anteroom.services.mcp_manager import _validate_sse_url

        with pytest.raises(ValueError, match="Blocked internal hostname"):
            _validate_sse_url("http://localhost/sse")

    def test_gcp_metadata_hostname_blocked(self) -> None:
        from anteroom.services.mcp_manager import _validate_sse_url

        with pytest.raises(ValueError, match="Blocked internal hostname"):
            _validate_sse_url("http://metadata.google.internal/sse")

    def test_loopback_ip_blocked(self) -> None:
        from anteroom.services.mcp_manager import _validate_sse_url

        with pytest.raises(ValueError, match="Blocked internal IP"):
            _validate_sse_url("http://127.0.0.1/sse")

    def test_private_ip_10_x_blocked(self) -> None:
        from anteroom.services.mcp_manager import _validate_sse_url

        with pytest.raises(ValueError, match="Blocked internal IP"):
            _validate_sse_url("http://10.0.0.1/sse")

    def test_private_ip_192_168_blocked(self) -> None:
        from anteroom.services.mcp_manager import _validate_sse_url

        with pytest.raises(ValueError, match="Blocked internal IP"):
            _validate_sse_url("http://192.168.1.1/sse")

    def test_private_ip_172_16_blocked(self) -> None:
        from anteroom.services.mcp_manager import _validate_sse_url

        with pytest.raises(ValueError, match="Blocked internal IP"):
            _validate_sse_url("http://172.16.0.1/sse")

    def test_link_local_ip_blocked(self) -> None:
        from anteroom.services.mcp_manager import _validate_sse_url

        with pytest.raises(ValueError, match="Blocked internal IP"):
            _validate_sse_url("http://169.254.169.254/sse")

    def test_ipv6_loopback_blocked(self) -> None:
        from anteroom.services.mcp_manager import _validate_sse_url

        with pytest.raises(ValueError, match="Blocked internal IP"):
            _validate_sse_url("http://[::1]/sse")

    def test_unresolvable_hostname_raises(self) -> None:
        from anteroom.services.mcp_manager import _validate_sse_url

        with pytest.raises(ValueError, match="Cannot resolve hostname"):
            _validate_sse_url("http://this-hostname-definitely-does-not-exist-xyzzy.invalid/sse")

    def test_hostname_resolving_to_private_ip_blocked(self) -> None:
        """Hostname that resolves to a private IP must be blocked (lines 51-57)."""
        import socket

        from anteroom.services.mcp_manager import _validate_sse_url

        private_ip = "10.0.0.1"
        with patch("anteroom.services.mcp_manager.socket.getaddrinfo") as mock_gai:
            mock_gai.return_value = [(socket.AF_INET, socket.SOCK_STREAM, 0, "", (private_ip, 0))]
            with pytest.raises(ValueError, match="resolves to blocked IP"):
                _validate_sse_url("http://internal-service.example.com/sse")


class TestValidateCommand:
    """Tests for _validate_command covering lines 64-66."""

    def test_existing_command_passes(self) -> None:
        from anteroom.services.mcp_manager import _validate_command

        with patch("anteroom.services.mcp_manager.shutil.which", return_value="/usr/bin/echo"):
            _validate_command("echo")  # should not raise

    def test_missing_command_raises(self) -> None:
        from anteroom.services.mcp_manager import _validate_command

        with patch("anteroom.services.mcp_manager.shutil.which", return_value=None):
            with pytest.raises(ValueError, match="MCP command not found on PATH"):
                _validate_command("nonexistent-mcp-tool")


class TestIsReadyProperty:
    """Tests for is_ready property covering lines 83-85."""

    def test_is_ready_true_when_no_configs(self) -> None:
        mgr = McpManager([])
        assert mgr.is_ready is True

    def test_is_ready_false_when_not_all_resolved(self) -> None:
        configs = [
            McpServerConfig(name="server-a", transport="stdio", command="echo"),
            McpServerConfig(name="server-b", transport="stdio", command="echo"),
        ]
        mgr = McpManager(configs)
        # Only one server has resolved status
        mgr._server_status = {"server-a": {"status": "connected", "tool_count": 0}}
        assert mgr.is_ready is False

    def test_is_ready_true_when_all_resolved(self) -> None:
        configs = [
            McpServerConfig(name="server-a", transport="stdio", command="echo"),
            McpServerConfig(name="server-b", transport="stdio", command="echo"),
        ]
        mgr = McpManager(configs)
        mgr._server_status = {
            "server-a": {"status": "connected", "tool_count": 2},
            "server-b": {"status": "error", "tool_count": 0},
        }
        assert mgr.is_ready is True


class TestStartupEmptyConfigs:
    """Tests for startup early-return with no configs (line 89)."""

    @pytest.mark.asyncio()
    async def test_startup_empty_configs_returns_immediately(self) -> None:
        mgr = McpManager([])
        # Patch MCP import to ensure we don't accidentally exercise connection code
        with patch.dict("sys.modules", {"mcp": MagicMock(), "mcp.client.stdio": MagicMock()}):
            await mgr.startup()
        assert mgr._server_status == {}
        assert mgr.is_ready is True


class TestStartupStatusCallback:
    """Tests for status_callback invocations during startup (lines 102-103, 182)."""

    @pytest.mark.asyncio()
    async def test_startup_calls_connecting_status_callback(self) -> None:
        """startup() must call status_callback with 'connecting' for each server (lines 102-103)."""
        configs = [
            McpServerConfig(name="server-a", transport="stdio", command="echo"),
            McpServerConfig(name="server-b", transport="stdio", command="echo"),
        ]
        mgr = McpManager(configs)
        received: list[tuple[str, dict]] = []

        async def fake_connect_one(config: McpServerConfig, status_callback: Any = None) -> None:
            mgr._server_status[config.name] = {"status": "connected", "tool_count": 0}

        with (
            patch.object(mgr, "_connect_one", side_effect=fake_connect_one),
            patch.dict("sys.modules", {"mcp": MagicMock(), "mcp.client.stdio": MagicMock()}),
        ):
            await mgr.startup(status_callback=lambda name, status: received.append((name, status)))

        connecting_calls = [(n, s) for n, s in received if s.get("status") == "connecting"]
        assert len(connecting_calls) == 2
        names = {n for n, _ in connecting_calls}
        assert names == {"server-a", "server-b"}

    @pytest.mark.asyncio()
    async def test_connect_one_calls_status_callback_on_completion(self) -> None:
        """_connect_one must call status_callback with final status (line 182)."""
        config = McpServerConfig(name="srv", transport="stdio", command="echo")
        mgr = McpManager([config])
        callback_calls: list[tuple[str, dict]] = []

        async def fake_do_connect(_config: McpServerConfig) -> None:
            mgr._server_status[_config.name] = {"status": "connected", "tool_count": 1}

        with patch.object(mgr, "_do_connect", side_effect=fake_do_connect):
            await mgr._connect_one(config, status_callback=lambda n, s: callback_calls.append((n, s)))

        assert len(callback_calls) == 1
        name, status = callback_calls[0]
        assert name == "srv"
        assert status["status"] == "connected"

    @pytest.mark.asyncio()
    async def test_connect_one_timeout_calls_status_callback(self) -> None:
        """status_callback must be called even when connection times out (line 182)."""
        config = McpServerConfig(name="slow", transport="stdio", command="echo", timeout=0.05)
        mgr = McpManager([config])
        callback_calls: list[tuple[str, dict]] = []

        async def hang_forever(_config: McpServerConfig) -> None:
            await asyncio.sleep(999)

        with patch.object(mgr, "_do_connect", side_effect=hang_forever):
            await mgr._connect_one(config, status_callback=lambda n, s: callback_calls.append((n, s)))

        assert len(callback_calls) == 1
        name, status = callback_calls[0]
        assert name == "slow"
        assert status["status"] == "error"


class TestDoConnectImportError:
    """Tests for _do_connect MCP SDK ImportError path (lines 191-199)."""

    @pytest.mark.asyncio()
    async def test_do_connect_without_mcp_sdk_sets_error_status(self) -> None:
        """_do_connect with no mcp package sets error status and returns (lines 191-199)."""
        config = McpServerConfig(name="no-sdk", transport="stdio", command="echo")
        mgr = McpManager([config])

        with patch.dict("sys.modules", {"mcp": None, "mcp.client.stdio": None}):
            await mgr._do_connect(config)

        assert mgr._server_status["no-sdk"]["status"] == "error"
        assert "not installed" in mgr._server_status["no-sdk"]["error_message"].lower()
        assert "no-sdk" not in mgr._sessions


class TestDoConnectSseTransport:
    """Tests for SSE transport branch in _do_connect (lines 231-250)."""

    @pytest.mark.asyncio()
    async def test_sse_client_import_error_sets_error_status(self) -> None:
        """When sse_client can't be imported, status must be set to error (lines 231-241)."""
        config = McpServerConfig(name="sse-server", transport="sse", url="https://93.184.216.34/sse")
        mgr = McpManager([config])

        mcp_mock = MagicMock()
        with patch.dict(
            "sys.modules",
            {"mcp": mcp_mock, "mcp.client.stdio": MagicMock(), "mcp.client.sse": None},
        ):
            await mgr._do_connect(config)

        assert mgr._server_status["sse-server"]["status"] == "error"
        assert "sse" in mgr._server_status["sse-server"]["error_message"].lower()

    @pytest.mark.asyncio()
    async def test_sse_url_validation_blocks_private_ip(self) -> None:
        """_do_connect for SSE must reject private IPs via _validate_sse_url (line 243)."""
        config = McpServerConfig(name="ssrf-attempt", transport="sse", url="https://mcp.example.com/sse")
        mgr = McpManager([config])

        mcp_mock = MagicMock()

        # Make McpError a real class so isinstance() works during error handling
        class FakeMcpError(Exception):
            pass

        mcp_mock.McpError = FakeMcpError
        sse_mod_mock = MagicMock()

        with (
            patch.dict(
                "sys.modules",
                {"mcp": mcp_mock, "mcp.client.stdio": MagicMock(), "mcp.client.sse": sse_mod_mock},
            ),
            patch(
                "anteroom.services.mcp_manager._validate_sse_url",
                side_effect=ValueError("Blocked internal IP: 10.0.0.1"),
            ),
        ):
            await mgr._do_connect(config)

        assert mgr._server_status["ssrf-attempt"]["status"] == "error"
        assert "Blocked internal IP" in mgr._server_status["ssrf-attempt"]["error_message"]

    @pytest.mark.asyncio()
    async def test_sse_connect_success(self) -> None:
        """SSE transport happy path: session initialized, tools listed (lines 244-250)."""
        config = McpServerConfig(name="sse-ok", transport="sse", url="https://93.184.216.34/sse")
        mgr = McpManager([config])

        mock_tool = MagicMock()
        mock_tool.name = "sse_tool"
        mock_tool.description = "An SSE tool"
        mock_tool.inputSchema = {"type": "object"}

        mock_session = AsyncMock()
        mock_tools_result = MagicMock()
        mock_tools_result.tools = [mock_tool]
        mock_session.initialize = AsyncMock()
        mock_session.list_tools = AsyncMock(return_value=mock_tools_result)

        mock_stack = AsyncMock(spec=AsyncExitStack)
        mock_stack.enter_async_context = AsyncMock(side_effect=[(MagicMock(), MagicMock()), mock_session])
        mock_stack.callback = MagicMock()
        mock_stack.aclose = AsyncMock()

        mcp_mock = MagicMock()
        sse_mod_mock = MagicMock()
        sse_mod_mock.sse_client = MagicMock()

        with (
            patch.dict(
                "sys.modules",
                {"mcp": mcp_mock, "mcp.client.stdio": MagicMock(), "mcp.client.sse": sse_mod_mock},
            ),
            patch("anteroom.services.mcp_manager.AsyncExitStack", return_value=mock_stack),
            patch("anteroom.services.mcp_manager._validate_sse_url"),
        ):
            await mgr._do_connect(config)

        assert mgr._server_status["sse-ok"]["status"] == "connected"
        assert mgr._server_tools["sse-ok"][0]["name"] == "sse_tool"


class TestDoConnectStdioSuccessPath:
    """Tests for successful stdio connect path (lines 224-228, 266-301)."""

    @pytest.mark.asyncio()
    async def test_stdio_connect_success_sets_connected_status(self) -> None:
        """Successful stdio connection populates session, tools, and connected status (lines 266-301)."""
        config = McpServerConfig(name="stdio-ok", transport="stdio", command="echo")
        mgr = McpManager([config])

        mock_tool = MagicMock()
        mock_tool.name = "my_tool"
        mock_tool.description = "Does things"
        mock_tool.inputSchema = {"type": "object", "properties": {}}

        mock_session = AsyncMock()
        mock_tools_result = MagicMock()
        mock_tools_result.tools = [mock_tool]
        mock_session.initialize = AsyncMock()
        mock_session.list_tools = AsyncMock(return_value=mock_tools_result)

        mock_stack = AsyncMock(spec=AsyncExitStack)
        # First enter_async_context returns (read, write) streams, second returns session
        mock_stack.enter_async_context = AsyncMock(side_effect=[(MagicMock(), MagicMock()), mock_session])
        mock_stack.callback = MagicMock()
        mock_stack.aclose = AsyncMock()

        mcp_mock = MagicMock()
        mcp_mock.ClientSession = MagicMock()
        mcp_mock.StdioServerParameters = MagicMock()

        with (
            patch.dict(
                "sys.modules",
                {"mcp": mcp_mock, "mcp.client.stdio": MagicMock()},
            ),
            patch("anteroom.services.mcp_manager.AsyncExitStack", return_value=mock_stack),
            patch("anteroom.services.mcp_manager.shutil.which", return_value="/usr/bin/echo"),
        ):
            await mgr._do_connect(config)

        assert mgr._server_status["stdio-ok"]["status"] == "connected"
        assert mgr._server_status["stdio-ok"]["tool_count"] == 1
        assert mgr._server_tools["stdio-ok"][0]["name"] == "my_tool"
        assert mgr._sessions["stdio-ok"] is mock_session

    @pytest.mark.asyncio()
    async def test_stdio_connect_with_filtered_tools_logs_filtered_count(self) -> None:
        """When tools are filtered, status reflects allowed count (lines 292-299)."""
        config = McpServerConfig(
            name="filtered-ok",
            transport="stdio",
            command="echo",
            tools_include=["allowed_tool"],
        )
        mgr = McpManager([config])

        mock_tool_allowed = MagicMock()
        mock_tool_allowed.name = "allowed_tool"
        mock_tool_allowed.description = "Allowed"
        mock_tool_allowed.inputSchema = {}

        mock_tool_blocked = MagicMock()
        mock_tool_blocked.name = "blocked_tool"
        mock_tool_blocked.description = "Blocked"
        mock_tool_blocked.inputSchema = {}

        mock_session = AsyncMock()
        mock_tools_result = MagicMock()
        mock_tools_result.tools = [mock_tool_allowed, mock_tool_blocked]
        mock_session.initialize = AsyncMock()
        mock_session.list_tools = AsyncMock(return_value=mock_tools_result)

        mock_stack = AsyncMock(spec=AsyncExitStack)
        mock_stack.enter_async_context = AsyncMock(side_effect=[(MagicMock(), MagicMock()), mock_session])
        mock_stack.callback = MagicMock()
        mock_stack.aclose = AsyncMock()

        mcp_mock = MagicMock()

        with (
            patch.dict("sys.modules", {"mcp": mcp_mock, "mcp.client.stdio": MagicMock()}),
            patch("anteroom.services.mcp_manager.AsyncExitStack", return_value=mock_stack),
            patch("anteroom.services.mcp_manager.shutil.which", return_value="/usr/bin/echo"),
        ):
            await mgr._do_connect(config)

        # 1 allowed out of 2 total
        assert mgr._server_status["filtered-ok"]["tool_count"] == 1
        assert mgr._server_status["filtered-ok"]["total_tool_count"] == 2

    @pytest.mark.asyncio()
    async def test_stdio_connect_tool_with_no_input_schema_attribute(self) -> None:
        """Tool without inputSchema attribute uses empty dict fallback (line 277)."""
        config = McpServerConfig(name="no-schema", transport="stdio", command="echo")
        mgr = McpManager([config])

        mock_tool = MagicMock(spec=["name", "description"])  # no inputSchema
        mock_tool.name = "simple_tool"
        mock_tool.description = "No schema"

        mock_session = AsyncMock()
        mock_tools_result = MagicMock()
        mock_tools_result.tools = [mock_tool]
        mock_session.initialize = AsyncMock()
        mock_session.list_tools = AsyncMock(return_value=mock_tools_result)

        mock_stack = AsyncMock(spec=AsyncExitStack)
        mock_stack.enter_async_context = AsyncMock(side_effect=[(MagicMock(), MagicMock()), mock_session])
        mock_stack.callback = MagicMock()
        mock_stack.aclose = AsyncMock()

        mcp_mock = MagicMock()

        with (
            patch.dict("sys.modules", {"mcp": mcp_mock, "mcp.client.stdio": MagicMock()}),
            patch("anteroom.services.mcp_manager.AsyncExitStack", return_value=mock_stack),
            patch("anteroom.services.mcp_manager.shutil.which", return_value="/usr/bin/echo"),
        ):
            await mgr._do_connect(config)

        assert mgr._server_tools["no-schema"][0]["input_schema"] == {}

    @pytest.mark.asyncio()
    async def test_stdio_connect_many_tools_truncates_log_names(self) -> None:
        """More than 10 tools should have '...' in the log (lines 298, 305)."""
        config = McpServerConfig(name="many-tools", transport="stdio", command="echo")
        mgr = McpManager([config])

        tools = []
        for i in range(15):
            t = MagicMock()
            t.name = f"tool_{i}"
            t.description = f"Tool {i}"
            t.inputSchema = {}
            tools.append(t)

        mock_session = AsyncMock()
        mock_tools_result = MagicMock()
        mock_tools_result.tools = tools
        mock_session.initialize = AsyncMock()
        mock_session.list_tools = AsyncMock(return_value=mock_tools_result)

        mock_stack = AsyncMock(spec=AsyncExitStack)
        mock_stack.enter_async_context = AsyncMock(side_effect=[(MagicMock(), MagicMock()), mock_session])
        mock_stack.callback = MagicMock()
        mock_stack.aclose = AsyncMock()

        mcp_mock = MagicMock()

        with (
            patch.dict("sys.modules", {"mcp": mcp_mock, "mcp.client.stdio": MagicMock()}),
            patch("anteroom.services.mcp_manager.AsyncExitStack", return_value=mock_stack),
            patch("anteroom.services.mcp_manager.shutil.which", return_value="/usr/bin/echo"),
        ):
            await mgr._do_connect(config)

        assert mgr._server_status["many-tools"]["tool_count"] == 15


class TestMcpErrorImportFailureDuringErrorHandling:
    """Tests for the McpError import fallback in error handling (lines 329-330)."""

    @pytest.mark.asyncio()
    async def test_error_handling_when_mcp_error_import_fails(self) -> None:
        """If 'from mcp import McpError' fails, error must still be recorded (lines 327-330)."""
        config = McpServerConfig(name="no-mcp-error", transport="stdio", command="echo")
        mgr = McpManager([config])

        # Patch shutil.which to find the command, then AsyncExitStack to raise on enter
        mock_stack = AsyncMock(spec=AsyncExitStack)
        mock_stack.enter_async_context = AsyncMock(side_effect=OSError("broken pipe"))
        mock_stack.aclose = AsyncMock()

        # mcp module exists for first import but McpError import attempt fails
        mcp_mock = MagicMock()
        del mcp_mock.McpError  # make McpError unavailable

        with (
            patch.dict("sys.modules", {"mcp": mcp_mock, "mcp.client.stdio": MagicMock()}),
            patch("anteroom.services.mcp_manager.AsyncExitStack", return_value=mock_stack),
            patch("anteroom.services.mcp_manager.shutil.which", return_value="/usr/bin/echo"),
        ):
            await mgr._do_connect(config)

        assert mgr._server_status["no-mcp-error"]["status"] == "error"
        assert "OSError" in mgr._server_status["no-mcp-error"]["error_message"]


class TestToolCollisionWarning:
    """Tests for tool name collision warning in _rebuild_tool_map (line 367)."""

    def test_tool_name_collision_logs_warning(self, caplog: pytest.LogCaptureFixture) -> None:
        import logging

        mgr = McpManager([])
        mgr._server_tools = {
            "server-a": [_make_tool("shared_tool", "server-a")],
            "server-b": [_make_tool("shared_tool", "server-b")],
        }

        with caplog.at_level(logging.WARNING, logger="anteroom.services.mcp_manager"):
            mgr._rebuild_tool_map()

        collision_records = [r for r in caplog.records if "collision" in r.message.lower()]
        assert collision_records, "Expected a tool name collision warning"
        assert "shared_tool" in collision_records[0].message

    def test_tool_collision_last_server_wins(self) -> None:
        mgr = McpManager([])
        # Server iteration order: server-a first, then server-b (dicts are insertion-ordered in Python 3.7+)
        mgr._server_tools = {
            "server-a": [_make_tool("shared_tool", "server-a")],
            "server-b": [_make_tool("shared_tool", "server-b")],
        }
        mgr._rebuild_tool_map()
        # server-b overwrites server-a
        assert mgr._tool_to_server["shared_tool"] == "server-b"


class TestConnectServerWithExistingSession:
    """Tests for connect_server reconnection path (lines 381-386)."""

    @pytest.mark.asyncio()
    async def test_connect_server_disconnects_first_if_already_connected(self) -> None:
        """connect_server must disconnect before reconnecting if session exists (lines 381-382)."""
        config = McpServerConfig(name="srv", transport="stdio", command="echo")
        mgr = McpManager([config])
        mgr._sessions["srv"] = AsyncMock()

        disconnect_called = False
        connect_called = False

        async def fake_disconnect(name: str) -> None:
            nonlocal disconnect_called
            disconnect_called = True
            mgr._sessions.pop(name, None)
            mgr._server_status[name] = {"status": "disconnected", "tool_count": 0}

        async def fake_connect_one(cfg: McpServerConfig, status_callback: Any = None) -> None:
            nonlocal connect_called
            connect_called = True
            mgr._server_status[cfg.name] = {"status": "connected", "tool_count": 0}

        with (
            patch.object(mgr, "disconnect_server", side_effect=fake_disconnect),
            patch.object(mgr, "_connect_one", side_effect=fake_connect_one),
        ):
            await mgr.connect_server("srv")

        assert disconnect_called, "disconnect_server should have been called first"
        assert connect_called, "_connect_one should have been called after disconnect"
        assert "srv" not in mgr._disabled

    @pytest.mark.asyncio()
    async def test_connect_server_no_existing_session_skips_disconnect(self) -> None:
        """connect_server without existing session should not call disconnect (line 381)."""
        config = McpServerConfig(name="fresh", transport="stdio", command="echo")
        mgr = McpManager([config])
        # No existing session

        disconnect_called = False

        async def fake_disconnect(name: str) -> None:
            nonlocal disconnect_called
            disconnect_called = True

        async def fake_connect_one(cfg: McpServerConfig, status_callback: Any = None) -> None:
            mgr._server_status[cfg.name] = {"status": "connected", "tool_count": 0}

        with (
            patch.object(mgr, "disconnect_server", side_effect=fake_disconnect),
            patch.object(mgr, "_connect_one", side_effect=fake_connect_one),
        ):
            await mgr.connect_server("fresh")

        assert not disconnect_called, "disconnect_server should NOT be called for a fresh server"

    @pytest.mark.asyncio()
    async def test_connect_server_clears_disabled_flag(self) -> None:
        """connect_server must clear the _disabled flag (line 384)."""
        config = McpServerConfig(name="re-enable", transport="stdio", command="echo")
        mgr = McpManager([config])
        mgr._disabled.add("re-enable")

        async def fake_connect_one(cfg: McpServerConfig, status_callback: Any = None) -> None:
            mgr._server_status[cfg.name] = {"status": "connected", "tool_count": 0}

        with patch.object(mgr, "_connect_one", side_effect=fake_connect_one):
            await mgr.connect_server("re-enable")

        assert "re-enable" not in mgr._disabled


class TestReconnectServer:
    """Tests for reconnect_server (line 413)."""

    @pytest.mark.asyncio()
    async def test_reconnect_server_delegates_to_connect_server(self) -> None:
        """reconnect_server is a thin wrapper around connect_server (line 413)."""
        config = McpServerConfig(name="srv", transport="stdio", command="echo")
        mgr = McpManager([config])

        connect_server_called_with: list[str] = []

        async def fake_connect_server(name: str) -> None:
            connect_server_called_with.append(name)

        with patch.object(mgr, "connect_server", side_effect=fake_connect_server):
            await mgr.reconnect_server("srv")

        assert connect_server_called_with == ["srv"]


class TestCallToolResultParsing:
    """Tests for call_tool result content parsing (lines 463-466, 473)."""

    @pytest.mark.asyncio()
    async def test_call_tool_result_with_data_attribute(self) -> None:
        """Items with .data (not .text) should be converted to str (line 463-464)."""
        mgr = McpManager([])
        mgr._tool_to_server = {"image_tool": "vision-server"}

        item = MagicMock(spec=["data"])  # has .data, no .text
        item.data = b"\x89PNG binary"

        mock_result = MagicMock()
        mock_result.content = [item]

        mock_session = AsyncMock()
        mock_session.call_tool = AsyncMock(return_value=mock_result)
        mgr._sessions = {"vision-server": mock_session}

        result = await mgr.call_tool("image_tool", {})
        assert "content" in result
        assert str(b"\x89PNG binary") in result["content"]

    @pytest.mark.asyncio()
    async def test_call_tool_result_with_plain_item(self) -> None:
        """Items with neither .text nor .data should fall through to str(item) (line 465-466)."""
        mgr = McpManager([])
        mgr._tool_to_server = {"plain_tool": "plain-server"}

        class PlainItem:
            def __str__(self) -> str:
                return "plain string result"

        mock_result = MagicMock()
        mock_result.content = [PlainItem()]

        mock_session = AsyncMock()
        mock_session.call_tool = AsyncMock(return_value=mock_result)
        mgr._sessions = {"plain-server": mock_session}

        result = await mgr.call_tool("plain_tool", {})
        assert "content" in result
        assert "plain string result" in result["content"]

    @pytest.mark.asyncio()
    async def test_call_tool_result_without_content_attribute(self) -> None:
        """Result without .content falls back to str(result) (line 473)."""
        mgr = McpManager([])
        mgr._tool_to_server = {"raw_tool": "raw-server"}

        class RawResult:
            def __str__(self) -> str:
                return "raw result string"

        mock_session = AsyncMock()
        mock_session.call_tool = AsyncMock(return_value=RawResult())
        mgr._sessions = {"raw-server": mock_session}

        result = await mgr.call_tool("raw_tool", {})
        assert "result" in result
        assert "raw result string" in result["result"]

    @pytest.mark.asyncio()
    async def test_call_tool_context_trust_from_config(self) -> None:
        """call_tool should include _context_trust from the server config (lines 455-456)."""
        config = McpServerConfig(name="trusted-server", transport="stdio", command="echo", trust_level="trusted")
        mgr = McpManager([config])
        mgr._tool_to_server = {"trusted_tool": "trusted-server"}

        mock_result = MagicMock()
        mock_result.content = [MagicMock(text="trusted response")]

        mock_session = AsyncMock()
        mock_session.call_tool = AsyncMock(return_value=mock_result)
        mgr._sessions = {"trusted-server": mock_session}

        result = await mgr.call_tool("trusted_tool", {})
        assert result["_context_trust"] == "trusted"
        assert result["_context_origin"] == "mcp:trusted-server"

    @pytest.mark.asyncio()
    async def test_call_tool_default_trust_level_when_no_config(self) -> None:
        """call_tool uses 'untrusted' when server config is missing (lines 455-456)."""
        mgr = McpManager([])
        mgr._tool_to_server = {"ghost_tool": "ghost-server"}
        # No config for ghost-server

        mock_result = MagicMock()
        mock_result.content = [MagicMock(text="response")]

        mock_session = AsyncMock()
        mock_session.call_tool = AsyncMock(return_value=mock_result)
        mgr._sessions = {"ghost-server": mock_session}

        result = await mgr.call_tool("ghost_tool", {})
        assert result["_context_trust"] == "untrusted"


class TestToolWarningThresholdWithFilters:
    """Tests for tool warning threshold with active filters (line 136)."""

    @pytest.mark.asyncio()
    async def test_tool_warning_with_active_filter_changes_hint(self, caplog: pytest.LogCaptureFixture) -> None:
        """When any server has an active filter and threshold is exceeded, the hint changes (line 136)."""
        import logging

        cfg = McpServerConfig(
            name="jira",
            transport="stdio",
            command="echo",
            tools_include=["tool0", "tool1", "tool2", "tool3", "tool4"],
        )
        mgr = McpManager([cfg], tool_warning_threshold=3)
        mgr._server_tools = {"jira": [_make_tool(f"tool{i}", "jira") for i in range(5)]}
        mgr._rebuild_tool_map()
        mgr._server_status = {"jira": {"status": "connected", "tool_count": 5}}

        async def noop(_config: McpServerConfig) -> None:
            pass

        with (
            caplog.at_level(logging.WARNING, logger="anteroom.services.mcp_manager"),
            patch.object(mgr, "_connect_one", side_effect=noop),
            patch.dict("sys.modules", {"mcp": MagicMock(), "mcp.client.stdio": MagicMock()}),
        ):
            await mgr.startup()

        warning_records = [r for r in caplog.records if "exceeds threshold" in r.message]
        assert warning_records, "Expected tool threshold warning"
        assert "not be restrictive enough" in warning_records[0].message



"""Tests for AIService retry logic, timeouts, and cancel handling (#1021)."""

from __future__ import annotations

import asyncio
from typing import Any
from unittest.mock import AsyncMock, MagicMock, patch

import httpx
import pytest

from anteroom.config import AIConfig
from anteroom.services.ai_service import AIService, _is_html_error


def _make_config(**overrides: Any) -> AIConfig:
    defaults: dict[str, Any] = {
        "base_url": "http://localhost:11434/v1",
        "api_key": "test-key",
        "model": "gpt-4",
        "request_timeout": 120,
        "verify_ssl": True,
        "retry_max_attempts": 0,
        "retry_backoff_base": 0.01,
        "first_token_timeout": 60,
        "chunk_stall_timeout": 45,
    }
    defaults.update(overrides)
    return AIConfig(**defaults)


def _make_service(**config_overrides: Any) -> AIService:
    config = _make_config(**config_overrides)
    svc = AIService.__new__(AIService)
    svc.config = config
    svc._token_provider = None
    svc.client = MagicMock()
    return svc


async def _collect(gen: Any) -> list[dict[str, Any]]:
    events: list[dict[str, Any]] = []
    async for e in gen:
        events.append(e)
    return events


class _MockChunk:
    def __init__(
        self,
        content: str | None = None,
        finish_reason: str | None = None,
        tool_calls: list[Any] | None = None,
        usage: Any = None,
    ) -> None:
        delta = MagicMock()
        delta.content = content
        delta.tool_calls = tool_calls
        choice = MagicMock()
        choice.delta = delta
        choice.finish_reason = finish_reason
        self.choices = [choice]
        self.usage = usage


class _AsyncChunkIter:
    def __init__(self, chunks: list[Any]) -> None:
        self._chunks = list(chunks)
        self._index = 0

    def __aiter__(self) -> "_AsyncChunkIter":
        return self

    async def __anext__(self) -> Any:
        if self._index >= len(self._chunks):
            raise StopAsyncIteration
        c = self._chunks[self._index]
        self._index += 1
        return c

    async def aclose(self) -> None:
        pass


# ---------------------------------------------------------------------------
# Retry with exponential backoff
# ---------------------------------------------------------------------------


class TestRetryBackoff:
    @pytest.mark.asyncio
    async def test_retry_on_timeout_then_success(self) -> None:
        from openai import APITimeoutError

        svc = _make_service(retry_max_attempts=1, retry_backoff_base=0.01, first_token_timeout=60)

        call_count = 0
        usage = MagicMock()
        usage.prompt_tokens = 10
        usage.completion_tokens = 5
        usage.total_tokens = 15

        async def fake_create(**kwargs: Any) -> Any:
            nonlocal call_count
            call_count += 1
            if call_count == 1:
                raise APITimeoutError(request=MagicMock())
            return _AsyncChunkIter([_MockChunk(content="hi"), _MockChunk(finish_reason="stop", usage=usage)])

        # Patch _build_client to re-apply our fake_create after client rebuild
        def patched_build() -> None:
            svc.client = MagicMock()
            svc.client.chat.completions.create = AsyncMock(side_effect=fake_create)

        svc._build_client = patched_build
        svc.client.chat.completions.create = AsyncMock(side_effect=fake_create)
        events = await _collect(svc.stream_chat([{"role": "user", "content": "hi"}]))

        retrying = [e for e in events if e["event"] == "retrying"]
        assert len(retrying) == 1
        assert retrying[0]["data"]["delay"] == pytest.approx(0.01, abs=0.001)
        assert any(e["event"] == "done" for e in events)

    @pytest.mark.asyncio
    async def test_exponential_backoff_delay(self) -> None:
        from openai import APITimeoutError

        svc = _make_service(retry_max_attempts=2, retry_backoff_base=0.01)

        def patched_build() -> None:
            svc.client = MagicMock()
            svc.client.chat.completions.create = AsyncMock(side_effect=APITimeoutError(request=MagicMock()))

        svc._build_client = patched_build
        svc.client.chat.completions.create = AsyncMock(side_effect=APITimeoutError(request=MagicMock()))
        events = await _collect(svc.stream_chat([{"role": "user", "content": "hi"}]))

        retrying = [e for e in events if e["event"] == "retrying"]
        assert len(retrying) == 2
        assert retrying[0]["data"]["delay"] == pytest.approx(0.01, abs=0.001)
        assert retrying[1]["data"]["delay"] == pytest.approx(0.02, abs=0.001)


# ---------------------------------------------------------------------------
# First-token timeout
# ---------------------------------------------------------------------------


class TestFirstTokenTimeout:
    @pytest.mark.asyncio
    async def test_first_token_timeout_yields_error(self) -> None:
        svc = _make_service(first_token_timeout=0.05, request_timeout=0.05)

        class _SlowStream:
            def __aiter__(self) -> "_SlowStream":
                return self

            async def __anext__(self) -> Any:
                await asyncio.sleep(10)
                raise StopAsyncIteration

            async def close(self) -> None:
                pass

        svc.client.chat.completions.create = AsyncMock(return_value=_SlowStream())
        events = await _collect(svc.stream_chat([{"role": "user", "content": "hi"}]))

        error = [e for e in events if e["event"] == "error"]
        assert len(error) == 1
        assert error[0]["data"]["code"] == "timeout"


# ---------------------------------------------------------------------------
# Cancel at various phases
# ---------------------------------------------------------------------------


class TestCancelAtPhases:
    @pytest.mark.asyncio
    async def test_cancel_before_create(self) -> None:
        svc = _make_service()
        cancel = asyncio.Event()
        cancel.set()

        events = await _collect(svc.stream_chat([{"role": "user", "content": "hi"}], cancel_event=cancel))
        assert events == []

    @pytest.mark.asyncio
    async def test_cancel_during_create(self) -> None:
        svc = _make_service()
        cancel = asyncio.Event()

        async def slow_create(**kwargs: Any) -> Any:
            cancel.set()
            await asyncio.sleep(10)

        svc.client.chat.completions.create = AsyncMock(side_effect=slow_create)
        events = await _collect(svc.stream_chat([{"role": "user", "content": "hi"}], cancel_event=cancel))
        assert not any(e["event"] == "error" for e in events)

    @pytest.mark.asyncio
    async def test_cancel_during_retry_backoff(self) -> None:
        from openai import APITimeoutError

        svc = _make_service(retry_max_attempts=2, retry_backoff_base=100)
        cancel = asyncio.Event()

        svc.client.chat.completions.create = AsyncMock(side_effect=APITimeoutError(request=MagicMock()))

        async def set_cancel_soon() -> None:
            await asyncio.sleep(0.05)
            cancel.set()

        task = asyncio.create_task(set_cancel_soon())
        events = await _collect(svc.stream_chat([{"role": "user", "content": "hi"}], cancel_event=cancel))
        await task

        assert not any(e["event"] == "error" for e in events)


# ---------------------------------------------------------------------------
# Token refresh on 401
# ---------------------------------------------------------------------------


class TestTokenRefresh:
    @pytest.mark.asyncio
    async def test_auth_error_triggers_refresh(self) -> None:
        from openai import AuthenticationError

        config = _make_config()
        mock_tp = MagicMock()
        mock_tp.get_token.return_value = "old-key"

        with patch("anteroom.services.ai_service.AsyncOpenAI"):
            svc = AIService(config, token_provider=mock_tp)

        usage = MagicMock()
        usage.prompt_tokens = 10
        usage.completion_tokens = 5
        usage.total_tokens = 15
        call_count = 0

        async def fake_create(**kwargs: Any) -> Any:
            nonlocal call_count
            call_count += 1
            if call_count == 1:
                raise AuthenticationError(
                    message="invalid key",
                    response=httpx.Response(401, request=httpx.Request("POST", "http://test")),
                    body=None,
                )
            return _AsyncChunkIter([_MockChunk(content="ok"), _MockChunk(finish_reason="stop", usage=usage)])

        def patched_build() -> None:
            svc.client = MagicMock()
            svc.client.chat.completions.create = AsyncMock(side_effect=fake_create)

        svc._build_client = patched_build
        svc.client.chat.completions.create = AsyncMock(side_effect=fake_create)
        events = await _collect(svc.stream_chat([{"role": "user", "content": "hi"}]))
        assert any(e["event"] == "done" for e in events)


# ---------------------------------------------------------------------------
# Tool argument delta accumulation
# ---------------------------------------------------------------------------


class TestToolArgsDelta:
    @pytest.mark.asyncio
    async def test_args_accumulate_across_chunks(self) -> None:
        svc = _make_service()

        tc1 = MagicMock()
        tc1.index = 0
        tc1.id = "call_1"
        tc1.function = MagicMock()
        tc1.function.name = "read_file"
        tc1.function.arguments = '{"pa'

        tc2 = MagicMock()
        tc2.index = 0
        tc2.id = None
        tc2.function = MagicMock()
        tc2.function.name = None
        tc2.function.arguments = 'th": "/tmp"}'

        chunk1 = MagicMock()
        chunk1.choices = [MagicMock()]
        chunk1.choices[0].delta = MagicMock()
        chunk1.choices[0].delta.content = None
        chunk1.choices[0].delta.tool_calls = [tc1]
        chunk1.choices[0].finish_reason = None
        chunk1.usage = None

        chunk2 = MagicMock()
        chunk2.choices = [MagicMock()]
        chunk2.choices[0].delta = MagicMock()
        chunk2.choices[0].delta.content = None
        chunk2.choices[0].delta.tool_calls = [tc2]
        chunk2.choices[0].finish_reason = None
        chunk2.usage = None

        chunk3 = MagicMock()
        chunk3.choices = [MagicMock()]
        chunk3.choices[0].delta = MagicMock()
        chunk3.choices[0].delta.content = None
        chunk3.choices[0].delta.tool_calls = None
        chunk3.choices[0].finish_reason = "tool_calls"
        chunk3.usage = None

        svc.client.chat.completions.create = AsyncMock(return_value=_AsyncChunkIter([chunk1, chunk2, chunk3]))
        events = await _collect(svc.stream_chat([{"role": "user", "content": "hi"}]))

        tc_events = [e for e in events if e["event"] == "tool_call"]
        assert len(tc_events) == 1
        assert tc_events[0]["data"]["arguments"] == {"path": "/tmp"}

        delta_events = [e for e in events if e["event"] == "tool_call_args_delta"]
        assert len(delta_events) == 2


# ---------------------------------------------------------------------------
# HTML error detection
# ---------------------------------------------------------------------------


class TestHtmlErrorDetection:
    def test_html_doctype(self) -> None:
        exc = Exception("<!DOCTYPE html><html>error page</html>")
        assert _is_html_error(exc) is True

    def test_html_tag(self) -> None:
        exc = Exception("<html>Something went wrong</html>")
        assert _is_html_error(exc) is True

    def test_json_error(self) -> None:
        exc = Exception('{"error": "something failed"}')
        assert _is_html_error(exc) is False

    @pytest.mark.asyncio
    async def test_html_error_yields_specific_message(self) -> None:
        svc = _make_service()
        svc.client.chat.completions.create = AsyncMock(
            side_effect=Exception("<!doctype html><html>Gateway Error</html>")
        )

        events = await _collect(svc.stream_chat([{"role": "user", "content": "hi"}]))
        err = [e for e in events if e["event"] == "error"]
        assert len(err) == 1
        assert err[0]["data"]["code"] == "html_response"
        assert "HTML" in err[0]["data"]["message"]


# ---------------------------------------------------------------------------
# Server error (5xx) with retry
# ---------------------------------------------------------------------------


class TestServerErrorRetry:
    @pytest.mark.asyncio
    async def test_500_retries_and_exhausts(self) -> None:
        from openai import APIStatusError

        svc = _make_service(retry_max_attempts=1, retry_backoff_base=0.01)
        resp = httpx.Response(500, request=httpx.Request("POST", "http://test"))
        exc = APIStatusError(message="server error", response=resp, body=None)

        def patched_build() -> None:
            svc.client = MagicMock()
            svc.client.chat.completions.create = AsyncMock(side_effect=exc)

        svc._build_client = patched_build
        svc.client.chat.completions.create = AsyncMock(side_effect=exc)
        events = await _collect(svc.stream_chat([{"role": "user", "content": "hi"}]))

        retrying = [e for e in events if e["event"] == "retrying"]
        assert len(retrying) == 1
        error = [e for e in events if e["event"] == "error"]
        assert len(error) == 1
        assert error[0]["data"]["retryable"] is True



"""Tests for /pack REPL command logic: service interactions and config writing (#525)."""

from __future__ import annotations

import sqlite3
import textwrap
from pathlib import Path

import pytest
import yaml

from anteroom.db import _FTS_SCHEMA, _FTS_TRIGGERS, _SCHEMA, ThreadSafeConnection
from anteroom.services.packs import (
    get_pack,
    install_pack,
    list_packs,
    parse_manifest,
    remove_pack,
    validate_manifest,
)


@pytest.fixture()
def db() -> ThreadSafeConnection:
    conn = sqlite3.connect(":memory:", check_same_thread=False)
    conn.row_factory = sqlite3.Row
    conn.execute("PRAGMA foreign_keys=ON")
    conn.executescript(_SCHEMA)
    try:
        conn.executescript(_FTS_SCHEMA)
        conn.executescript(_FTS_TRIGGERS)
    except sqlite3.OperationalError:
        pass
    conn.commit()
    return ThreadSafeConnection(conn)


@pytest.fixture()
def pack_dir(tmp_path: Path) -> Path:
    """Create a minimal valid pack directory with a manifest and one skill."""
    d = tmp_path / "test-pack"
    d.mkdir()
    skill_dir = d / "skills"
    skill_dir.mkdir()
    (skill_dir / "greet.yaml").write_text(
        textwrap.dedent("""\
            name: greet
            description: Say hello
            prompt: Hello!
        """)
    )
    (d / "pack.yaml").write_text(
        textwrap.dedent("""\
            name: test-pack
            namespace: myteam
            version: "1.0.0"
            description: A test pack
            artifacts:
              - type: skill
                name: greet
                file: skills/greet.yaml
        """)
    )
    return d


# ---------------------------------------------------------------------------
# /pack list
# ---------------------------------------------------------------------------


class TestPackList:
    def test_empty_when_no_packs(self, db: ThreadSafeConnection) -> None:
        assert list_packs(db) == []

    def test_shows_installed_packs(self, db: ThreadSafeConnection, pack_dir: Path) -> None:
        manifest = parse_manifest(pack_dir / "pack.yaml")
        install_pack(db, manifest, pack_dir)
        result = list_packs(db)
        assert len(result) == 1
        assert result[0]["name"] == "test-pack"
        assert result[0]["namespace"] == "myteam"
        assert result[0]["artifact_count"] == 1


# ---------------------------------------------------------------------------
# /pack show
# ---------------------------------------------------------------------------


class TestPackShow:
    def test_returns_none_for_missing(self, db: ThreadSafeConnection) -> None:
        assert get_pack(db, "myteam", "nonexistent") is None

    def test_returns_pack_with_artifacts(self, db: ThreadSafeConnection, pack_dir: Path) -> None:
        manifest = parse_manifest(pack_dir / "pack.yaml")
        install_pack(db, manifest, pack_dir)
        result = get_pack(db, "myteam", "test-pack")
        assert result is not None
        assert result["name"] == "test-pack"
        assert len(result.get("artifacts", [])) == 1


# ---------------------------------------------------------------------------
# /pack install
# ---------------------------------------------------------------------------


class TestPackInstall:
    def test_install_from_dir(self, db: ThreadSafeConnection, pack_dir: Path) -> None:
        manifest = parse_manifest(pack_dir / "pack.yaml")
        errors = validate_manifest(manifest, pack_dir)
        assert errors == []
        result = install_pack(db, manifest, pack_dir)
        assert result["name"] == "test-pack"
        assert result["artifact_count"] == 1

    def test_install_duplicate_allowed(self, db: ThreadSafeConnection, pack_dir: Path) -> None:
        manifest = parse_manifest(pack_dir / "pack.yaml")
        r1 = install_pack(db, manifest, pack_dir)
        r2 = install_pack(db, manifest, pack_dir)
        assert r1["id"] != r2["id"]

    def test_parse_manifest_missing_file(self, tmp_path: Path) -> None:
        with pytest.raises((ValueError, FileNotFoundError)):
            parse_manifest(tmp_path / "nonexistent" / "pack.yaml")


# ---------------------------------------------------------------------------
# /pack remove
# ---------------------------------------------------------------------------


class TestPackRemove:
    def test_remove_installed(self, db: ThreadSafeConnection, pack_dir: Path) -> None:
        manifest = parse_manifest(pack_dir / "pack.yaml")
        install_pack(db, manifest, pack_dir)
        assert remove_pack(db, "myteam", "test-pack") is True
        assert list_packs(db) == []

    def test_remove_nonexistent(self, db: ThreadSafeConnection) -> None:
        assert remove_pack(db, "myteam", "nonexistent") is False


# ---------------------------------------------------------------------------
# /pack add-source (config write)
# ---------------------------------------------------------------------------


class TestPackAddSourceValidation:
    """URL scheme validation — security gate before config write."""

    def test_rejects_ext_scheme(self) -> None:
        from anteroom.services.pack_sources import _validate_url_scheme

        err = _validate_url_scheme("ext::sh -c 'echo pwned'")
        assert err is not None
        assert "not allowed" in err

    def test_rejects_file_scheme(self) -> None:
        from anteroom.services.pack_sources import _validate_url_scheme

        err = _validate_url_scheme("file:///etc/passwd")
        assert err is not None
        assert "not allowed" in err

    def test_allows_https(self) -> None:
        from anteroom.services.pack_sources import _validate_url_scheme

        assert _validate_url_scheme("https://github.com/org/repo.git") is None

    def test_allows_ssh_shorthand(self) -> None:
        from anteroom.services.pack_sources import _validate_url_scheme

        assert _validate_url_scheme("git@github.com:org/repo.git") is None

    def test_http_rejected_by_validator(self) -> None:
        """_validate_url_scheme rejects http:// to prevent MITM attacks."""
        from anteroom.services.pack_sources import _validate_url_scheme

        result = _validate_url_scheme("http://example.com/repo.git")
        assert result is not None
        assert "HTTP" in result


class TestPackAddSource:
    def test_adds_source_to_config(self, tmp_path: Path) -> None:
        config_path = tmp_path / "config.yaml"
        config_path.write_text("ai:\n  model: gpt-4o\n")

        raw: dict = yaml.safe_load(config_path.read_text()) or {}
        sources_list: list = raw.setdefault("pack_sources", [])
        sources_list.append({"url": "https://git.example.com/packs.git", "branch": "main", "refresh_interval": 30})
        with open(config_path, "w", encoding="utf-8") as f:
            yaml.dump(raw, f, default_flow_style=False, sort_keys=False)

        reloaded = yaml.safe_load(config_path.read_text())
        assert len(reloaded["pack_sources"]) == 1
        assert reloaded["pack_sources"][0]["url"] == "https://git.example.com/packs.git"
        assert reloaded["ai"]["model"] == "gpt-4o"

    def test_does_not_duplicate(self, tmp_path: Path) -> None:
        config_path = tmp_path / "config.yaml"
        config_path.write_text(
            "pack_sources:\n- url: https://git.example.com/packs.git\n  branch: main\n  refresh_interval: 30\n"
        )

        raw: dict = yaml.safe_load(config_path.read_text()) or {}
        sources_list: list = raw.setdefault("pack_sources", [])
        existing_urls = [s.get("url") for s in sources_list if isinstance(s, dict)]
        url = "https://git.example.com/packs.git"
        assert url in existing_urls

    def test_creates_config_if_missing(self, tmp_path: Path) -> None:
        config_path = tmp_path / "config.yaml"

        raw: dict = {}
        sources_list: list = raw.setdefault("pack_sources", [])
        sources_list.append({"url": "https://bb.example.com/repo.git", "branch": "main", "refresh_interval": 30})
        config_path.parent.mkdir(parents=True, exist_ok=True)
        with open(config_path, "w", encoding="utf-8") as f:
            yaml.dump(raw, f, default_flow_style=False, sort_keys=False)

        reloaded = yaml.safe_load(config_path.read_text())
        assert len(reloaded["pack_sources"]) == 1


# ---------------------------------------------------------------------------
# /pack sources — list_cached_sources
# ---------------------------------------------------------------------------


class TestPackSources:
    def test_empty_when_no_cache(self, tmp_path: Path) -> None:
        from anteroom.services.pack_sources import list_cached_sources

        assert list_cached_sources(tmp_path) == []


# ---------------------------------------------------------------------------
# /pack attach + /pack detach — service-level tests (#559)
# ---------------------------------------------------------------------------


class TestPackAttachRepl:
    def test_attach_global(self, db: ThreadSafeConnection, pack_dir: Path) -> None:
        from anteroom.services.pack_attachments import attach_pack, list_attachments, resolve_pack_id

        manifest = parse_manifest(pack_dir / "pack.yaml")
        install_pack(db, manifest, pack_dir)
        pack_id = resolve_pack_id(db, "myteam", "test-pack")
        assert pack_id is not None
        attach_pack(db, pack_id, project_path=None)
        attachments = list_attachments(db, project_path=None)
        assert len(attachments) == 1
        assert attachments[0]["scope"] == "global"

    def test_attach_project(self, db: ThreadSafeConnection, pack_dir: Path) -> None:
        from anteroom.services.pack_attachments import attach_pack, list_attachments, resolve_pack_id

        manifest = parse_manifest(pack_dir / "pack.yaml")
        install_pack(db, manifest, pack_dir)
        pack_id = resolve_pack_id(db, "myteam", "test-pack")
        assert pack_id is not None
        attach_pack(db, pack_id, project_path="/tmp/my-project")
        attachments = list_attachments(db, project_path="/tmp/my-project")
        assert any(a["scope"] == "project" for a in attachments)

    def test_attach_not_found(self, db: ThreadSafeConnection) -> None:
        from anteroom.services.pack_attachments import resolve_pack_id

        assert resolve_pack_id(db, "no", "such-pack") is None

    def test_attach_already_attached(self, db: ThreadSafeConnection, pack_dir: Path) -> None:
        from anteroom.services.pack_attachments import attach_pack, resolve_pack_id

        manifest = parse_manifest(pack_dir / "pack.yaml")
        install_pack(db, manifest, pack_dir)
        pack_id = resolve_pack_id(db, "myteam", "test-pack")
        assert pack_id is not None
        attach_pack(db, pack_id, project_path=None)
        with pytest.raises(ValueError, match="already attached"):
            attach_pack(db, pack_id, project_path=None)


class TestPackDetachRepl:
    def test_detach_success(self, db: ThreadSafeConnection, pack_dir: Path) -> None:
        from anteroom.services.pack_attachments import attach_pack, detach_pack, resolve_pack_id

        manifest = parse_manifest(pack_dir / "pack.yaml")
        install_pack(db, manifest, pack_dir)
        pack_id = resolve_pack_id(db, "myteam", "test-pack")
        assert pack_id is not None
        attach_pack(db, pack_id, project_path=None)
        assert detach_pack(db, pack_id, project_path=None) is True

    def test_detach_not_attached(self, db: ThreadSafeConnection, pack_dir: Path) -> None:
        from anteroom.services.pack_attachments import detach_pack, resolve_pack_id

        manifest = parse_manifest(pack_dir / "pack.yaml")
        install_pack(db, manifest, pack_dir)
        pack_id = resolve_pack_id(db, "myteam", "test-pack")
        assert pack_id is not None
        assert detach_pack(db, pack_id, project_path=None) is False

    def test_detach_project(self, db: ThreadSafeConnection, pack_dir: Path) -> None:
        from anteroom.services.pack_attachments import attach_pack, detach_pack, resolve_pack_id

        manifest = parse_manifest(pack_dir / "pack.yaml")
        install_pack(db, manifest, pack_dir)
        pack_id = resolve_pack_id(db, "myteam", "test-pack")
        assert pack_id is not None
        attach_pack(db, pack_id, project_path="/tmp/proj")
        assert detach_pack(db, pack_id, project_path="/tmp/proj") is True


# ---------------------------------------------------------------------------
# /pack update — service-level tests (#559)
# ---------------------------------------------------------------------------


class TestPackUpdateRepl:
    def test_update_success(self, db: ThreadSafeConnection, pack_dir: Path) -> None:
        from anteroom.services.packs import update_pack

        manifest = parse_manifest(pack_dir / "pack.yaml")
        install_pack(db, manifest, pack_dir)
        result = update_pack(db, manifest, pack_dir)
        assert result["name"] == "test-pack"
        assert result["namespace"] == "myteam"

    def test_update_missing_pack_yaml(self, tmp_path: Path) -> None:
        with pytest.raises(ValueError):
            parse_manifest(tmp_path / "nonexistent" / "pack.yaml")


# ---------------------------------------------------------------------------
# Built-in command registration
# ---------------------------------------------------------------------------


class TestPackCommandRegistration:
    def test_pack_in_builtin_commands(self) -> None:
        from anteroom.cli.skills import _BUILTIN_COMMANDS

        assert "pack" in _BUILTIN_COMMANDS
        assert "packs" in _BUILTIN_COMMANDS



"""Tests for the OpenAI-compatible proxy router."""

from __future__ import annotations

from unittest.mock import AsyncMock, MagicMock

from fastapi import FastAPI
from fastapi.testclient import TestClient

from anteroom.routers.proxy import router


def _make_app(*, mock_service: MagicMock | None = None) -> FastAPI:
    """Create a minimal FastAPI app with the proxy router."""
    app = FastAPI()
    app.include_router(router, prefix="/v1")

    config = MagicMock()
    config.ai.model = "gpt-4"
    app.state.config = config
    app.state.proxy_ai_service = mock_service

    return app


def _make_mock_service(*, response: MagicMock | None = None, error: Exception | None = None) -> MagicMock:
    """Create a mock AIService with pre-configured completions behavior."""
    service = MagicMock()
    if error:
        service.client.chat.completions.create = AsyncMock(side_effect=error)
    elif response:
        service.client.chat.completions.create = AsyncMock(return_value=response)
    return service


class TestListModels:
    def test_returns_configured_model(self) -> None:
        app = _make_app()
        client = TestClient(app)
        resp = client.get("/v1/models")
        assert resp.status_code == 200
        data = resp.json()
        assert data["object"] == "list"
        assert len(data["data"]) == 1
        assert data["data"][0]["id"] == "gpt-4"
        assert data["data"][0]["object"] == "model"


class TestChatCompletions:
    def test_rejects_non_json_content_type(self) -> None:
        app = _make_app()
        client = TestClient(app)
        resp = client.post("/v1/chat/completions", content="hello", headers={"content-type": "text/plain"})
        assert resp.status_code == 415

    def test_rejects_invalid_json(self) -> None:
        app = _make_app()
        client = TestClient(app)
        resp = client.post(
            "/v1/chat/completions",
            content="not json",
            headers={"content-type": "application/json"},
        )
        assert resp.status_code == 400

    def test_rejects_missing_messages(self) -> None:
        app = _make_app()
        client = TestClient(app)
        resp = client.post(
            "/v1/chat/completions",
            json={"model": "gpt-4"},
        )
        assert resp.status_code == 400
        assert "messages" in resp.json()["error"]["message"]

    def test_rejects_empty_messages(self) -> None:
        app = _make_app()
        client = TestClient(app)
        resp = client.post(
            "/v1/chat/completions",
            json={"model": "gpt-4", "messages": []},
        )
        assert resp.status_code == 400

    def test_rejects_non_object_body(self) -> None:
        app = _make_app()
        client = TestClient(app)
        resp = client.post(
            "/v1/chat/completions",
            json=["not", "an", "object"],
        )
        assert resp.status_code == 400
        assert "JSON object" in resp.json()["error"]["message"]

    def test_rejects_message_without_role(self) -> None:
        app = _make_app()
        client = TestClient(app)
        resp = client.post(
            "/v1/chat/completions",
            json={"messages": [{"content": "Hi"}]},
        )
        assert resp.status_code == 400
        assert "role" in resp.json()["error"]["message"]

    def test_rejects_non_dict_message_item(self) -> None:
        app = _make_app()
        client = TestClient(app)
        resp = client.post(
            "/v1/chat/completions",
            json={"messages": ["not a dict"]},
        )
        assert resp.status_code == 400
        assert "messages[0]" in resp.json()["error"]["message"]

    def test_non_streaming_success(self) -> None:
        mock_response = MagicMock()
        mock_response.model_dump.return_value = {
            "id": "chatcmpl-123",
            "object": "chat.completion",
            "choices": [{"message": {"role": "assistant", "content": "Hello!"}, "index": 0}],
        }
        service = _make_mock_service(response=mock_response)
        app = _make_app(mock_service=service)

        client = TestClient(app)
        resp = client.post(
            "/v1/chat/completions",
            json={"messages": [{"role": "user", "content": "Hi"}]},
        )
        assert resp.status_code == 200
        data = resp.json()
        assert data["id"] == "chatcmpl-123"
        assert data["choices"][0]["message"]["content"] == "Hello!"

    def test_uses_configured_model_as_default(self) -> None:
        mock_response = MagicMock()
        mock_response.model_dump.return_value = {"id": "chatcmpl-123", "choices": []}
        service = _make_mock_service(response=mock_response)
        app = _make_app(mock_service=service)

        client = TestClient(app)
        client.post(
            "/v1/chat/completions",
            json={"messages": [{"role": "user", "content": "Hi"}]},
        )

        call_kwargs = service.client.chat.completions.create.call_args[1]
        assert call_kwargs["model"] == "gpt-4"

    def test_allows_model_override(self) -> None:
        mock_response = MagicMock()
        mock_response.model_dump.return_value = {"id": "chatcmpl-123", "choices": []}
        service = _make_mock_service(response=mock_response)
        app = _make_app(mock_service=service)

        client = TestClient(app)
        client.post(
            "/v1/chat/completions",
            json={"model": "gpt-3.5-turbo", "messages": [{"role": "user", "content": "Hi"}]},
        )

        call_kwargs = service.client.chat.completions.create.call_args[1]
        assert call_kwargs["model"] == "gpt-3.5-turbo"

    def test_forwards_optional_parameters(self) -> None:
        mock_response = MagicMock()
        mock_response.model_dump.return_value = {"id": "chatcmpl-123", "choices": []}
        service = _make_mock_service(response=mock_response)
        app = _make_app(mock_service=service)

        client = TestClient(app)
        client.post(
            "/v1/chat/completions",
            json={
                "messages": [{"role": "user", "content": "Hi"}],
                "temperature": 0.7,
                "max_tokens": 100,
                "top_p": 0.9,
            },
        )

        call_kwargs = service.client.chat.completions.create.call_args[1]
        assert call_kwargs["temperature"] == 0.7
        assert call_kwargs["max_tokens"] == 100
        assert call_kwargs["top_p"] == 0.9

    def test_upstream_error_returns_502(self) -> None:
        service = _make_mock_service(error=Exception("upstream down"))
        app = _make_app(mock_service=service)

        client = TestClient(app)
        resp = client.post(
            "/v1/chat/completions",
            json={"messages": [{"role": "user", "content": "Hi"}]},
        )
        assert resp.status_code == 502
        assert resp.json()["error"]["message"] == "Upstream API error"

    def test_streaming_returns_sse(self) -> None:
        chunk1 = MagicMock()
        chunk1.model_dump.return_value = {
            "id": "chatcmpl-123",
            "object": "chat.completion.chunk",
            "choices": [{"delta": {"content": "Hi"}, "index": 0}],
        }

        async def mock_stream():
            yield chunk1

        service = MagicMock()
        service.client.chat.completions.create = AsyncMock(return_value=mock_stream())
        app = _make_app(mock_service=service)

        client = TestClient(app)
        resp = client.post(
            "/v1/chat/completions",
            json={"messages": [{"role": "user", "content": "Hi"}], "stream": True},
        )
        assert resp.status_code == 200
        assert resp.headers["content-type"].startswith("text/event-stream")
        body = resp.text
        assert "data: " in body
        assert "chatcmpl-123" in body
        assert "data: [DONE]" in body

    def test_streaming_error_emits_error_chunk(self) -> None:
        async def failing_stream():
            raise RuntimeError("connection lost")
            yield  # noqa: F811 — unreachable but needed to make this an async generator

        service = MagicMock()
        service.client.chat.completions.create = AsyncMock(return_value=failing_stream())
        app = _make_app(mock_service=service)

        client = TestClient(app)
        resp = client.post(
            "/v1/chat/completions",
            json={"messages": [{"role": "user", "content": "Hi"}], "stream": True},
        )
        assert resp.status_code == 200
        body = resp.text
        assert "Upstream stream error" in body
        assert "data: [DONE]" in body


class TestProxyConfig:
    def test_proxy_config_defaults(self) -> None:
        from anteroom.config import ProxyConfig

        config = ProxyConfig()
        assert config.enabled is False
        assert config.allowed_origins == []

    def test_proxy_config_in_app_config(self) -> None:
        from anteroom.config import AIConfig, AppConfig, ProxyConfig

        ai = AIConfig(base_url="http://localhost:8080/v1", api_key="test")
        app_config = AppConfig(ai=ai, proxy=ProxyConfig(enabled=True))
        assert app_config.proxy.enabled is True

    def test_proxy_config_parsed_from_yaml(self) -> None:
        import tempfile
        from pathlib import Path

        from anteroom.config import load_config

        yaml_content = """
ai:
  base_url: http://localhost:8080/v1
  api_key: test-key
proxy:
  enabled: true
  allowed_origins:
    - http://localhost:3000
"""
        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            f.write(yaml_content)
            f.flush()
            config, _ = load_config(Path(f.name))

        assert config.proxy.enabled is True
        assert config.proxy.allowed_origins == ["http://localhost:3000"]

    def test_proxy_config_env_override(self) -> None:
        import os
        import tempfile
        from pathlib import Path

        from anteroom.config import load_config

        yaml_content = """
ai:
  base_url: http://localhost:8080/v1
  api_key: test-key
"""
        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            f.write(yaml_content)
            f.flush()
            old = os.environ.get("AI_CHAT_PROXY_ENABLED")
            try:
                os.environ["AI_CHAT_PROXY_ENABLED"] = "true"
                config, _ = load_config(Path(f.name))
            finally:
                if old is None:
                    os.environ.pop("AI_CHAT_PROXY_ENABLED", None)
                else:
                    os.environ["AI_CHAT_PROXY_ENABLED"] = old

        assert config.proxy.enabled is True

    def test_proxy_config_rejects_wildcard_origin(self) -> None:
        import tempfile
        from pathlib import Path

        from anteroom.config import load_config

        yaml_content = """
ai:
  base_url: http://localhost:8080/v1
  api_key: test-key
proxy:
  enabled: true
  allowed_origins:
    - "*"
    - http://localhost:3000
    - not-a-url
"""
        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            f.write(yaml_content)
            f.flush()
            config, _ = load_config(Path(f.name))

        assert config.proxy.allowed_origins == ["http://localhost:3000"]



"""CLI integration tests for `aroom workflow` subcommands.

Tests real argparse dispatch via subprocess. Exercises the actual
`python -m anteroom workflow` command path — not direct handler calls.
"""

from __future__ import annotations

import subprocess
import sys
from pathlib import Path

import pytest

_PYTHON = sys.executable


@pytest.fixture(autouse=True)
def _isolated_home(tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
    """Use an isolated HOME so workflow runs don't pollute the user's DB."""
    anteroom_dir = tmp_path / ".anteroom"
    anteroom_dir.mkdir()
    config_file = anteroom_dir / "config.yaml"
    config_file.write_text('ai:\n  base_url: "http://localhost:1/v1"\n  api_key: "test"\n  model: "test"\n')
    monkeypatch.setenv("HOME", str(tmp_path))
    yield


# A generic test workflow (shell-only, no AI needed, no GitHub concepts)
# Uses no required inputs so it can be run via `aroom workflow run <path>`
# without needing --issue or other flags.
_TEST_WORKFLOW_YAML = """\
kind: workflow
id: test_cli_pipeline
version: 0.1.0
inputs: {}
steps:
  - id: greet
    type: runner
    runner: shell
    command: "echo Hello from workflow"
    timeout: 10
  - id: validate
    type: runner
    runner: shell
    command: "echo Validation passed"
    timeout: 10
"""


def _run_aroom(*args: str, timeout: int = 30) -> subprocess.CompletedProcess:
    """Run `aroom` CLI via subprocess and capture output."""
    return subprocess.run(
        [_PYTHON, "-m", "anteroom", *args],
        capture_output=True,
        text=True,
        timeout=timeout,
    )


# ---------------------------------------------------------------------------
# Help and error handling (subprocess)
# ---------------------------------------------------------------------------


class TestWorkflowCLIHelp:
    """Help text renders correctly for workflow subcommands."""

    def test_workflow_help(self) -> None:
        result = _run_aroom("workflow", "--help")
        assert result.returncode == 0
        assert "run" in result.stdout
        assert "status" in result.stdout
        assert "list" in result.stdout
        assert "history" in result.stdout

    def test_workflow_run_help(self) -> None:
        result = _run_aroom("workflow", "run", "--help")
        assert result.returncode == 0
        assert "workflow_name" in result.stdout
        assert "--issue" in result.stdout
        assert "--dry-run" in result.stdout
        assert "--detach" in result.stdout

    def test_workflow_status_help(self) -> None:
        result = _run_aroom("workflow", "status", "--help")
        assert result.returncode == 0
        assert "run_id" in result.stdout

    def test_workflow_list_help(self) -> None:
        result = _run_aroom("workflow", "list", "--help")
        assert result.returncode == 0
        assert "--status" in result.stdout
        assert "--limit" in result.stdout

    def test_workflow_history_help(self) -> None:
        result = _run_aroom("workflow", "history", "--help")
        assert result.returncode == 0
        assert "run_id" in result.stdout


class TestWorkflowCLIDryRun:
    """Dry run shows workflow plan without executing."""

    def test_dry_run_issue_delivery(self) -> None:
        result = _run_aroom("workflow", "run", "issue_delivery", "--dry-run", "--issue", "42")
        assert result.returncode == 0
        assert "issue_delivery" in result.stdout
        assert "v0.1.0" in result.stdout
        assert "issue:42" in result.stdout
        assert "gate_issue_current" in result.stdout
        assert "gate_plan" in result.stdout

    def test_dry_run_unknown_workflow(self) -> None:
        result = _run_aroom("workflow", "run", "nonexistent", "--dry-run")
        assert result.returncode == 0
        assert "Unknown workflow" in result.stdout or "Error" in result.stdout


# ---------------------------------------------------------------------------
# Real command path tests — run/status/list/history via subprocess
# ---------------------------------------------------------------------------


class TestWorkflowCLIRealCommands:
    """Drive real `python -m anteroom workflow` commands through subprocess.

    Uses a shell-only test workflow (no AI needed) written to a temp file.
    The actual argparse → config → handler → engine → storage path is
    exercised end-to-end.
    """

    @pytest.fixture()
    def workflow_file(self, tmp_path: Path) -> Path:
        """Write the test workflow YAML to a temp file."""
        wf = tmp_path / "test_pipeline.yaml"
        wf.write_text(_TEST_WORKFLOW_YAML)
        return wf

    def test_run_shell_workflow_completes(self, workflow_file: Path) -> None:
        """Run a shell-only workflow via the real CLI and verify completion."""
        result = _run_aroom(
            "workflow",
            "run",
            str(workflow_file),
            timeout=30,
        )
        # The workflow should complete (shell commands are simple echos)
        output = result.stdout + result.stderr
        assert "completed" in output.lower() or "Workflow completed" in output
        assert "Run ID:" in output

    def test_run_then_list_shows_run(self, workflow_file: Path) -> None:
        """After running a workflow, `list` shows it."""
        # Run the workflow first
        _run_aroom("workflow", "run", str(workflow_file), timeout=30)
        # List should show the run
        list_result = _run_aroom("workflow", "list")
        list_output = list_result.stdout
        assert list_result.returncode == 0
        # Should show either the workflow ID or the run table
        assert "test_cli_pipeline" in list_output or "Workflow Runs" in list_output
        assert "State" in list_output
        assert "Step" not in list_output

    def test_run_then_status_shows_details(self, workflow_file: Path) -> None:
        """After running, `status <run_id>` shows run details."""
        # Run the workflow and extract run ID
        run_result = _run_aroom(
            "workflow",
            "run",
            str(workflow_file),
            timeout=30,
        )
        run_output = run_result.stdout + run_result.stderr

        # Extract run ID from output
        run_id = None
        for line in run_output.splitlines():
            if "Run ID:" in line:
                run_id = line.split("Run ID:")[-1].strip()
                break

        if run_id:
            status_result = _run_aroom("workflow", "status", run_id)
            status_output = status_result.stdout
            assert status_result.returncode == 0
            assert "test_cli_pipeline" in status_output or run_id[:12] in status_output

    def test_run_then_history_shows_steps(self, workflow_file: Path) -> None:
        """After running, `history <run_id>` shows step details."""
        # Run the workflow and extract run ID
        run_result = _run_aroom(
            "workflow",
            "run",
            str(workflow_file),
            timeout=30,
        )
        run_output = run_result.stdout + run_result.stderr

        # Extract run ID
        run_id = None
        for line in run_output.splitlines():
            if "Run ID:" in line:
                run_id = line.split("Run ID:")[-1].strip()
                break

        if run_id:
            history_result = _run_aroom("workflow", "history", run_id)
            history_output = history_result.stdout
            assert history_result.returncode == 0
            # Should show step names from the workflow
            assert "greet" in history_output or "validate" in history_output

    def test_status_nonexistent_run(self) -> None:
        """Status with a fake run ID shows error, doesn't crash."""
        result = _run_aroom("workflow", "status", "nonexistent-fake-id")
        output = result.stdout + result.stderr
        assert "not found" in output.lower() or "Error" in output

    def test_list_empty_state(self) -> None:
        """List with no runs shows empty message."""
        result = _run_aroom("workflow", "list")
        assert result.returncode == 0
        output = result.stdout
        assert "No workflow runs" in output or "Workflow Runs" in output

    def test_list_output_is_run_centric(self, workflow_file: Path) -> None:
        """List view shows run-level columns, not step-level detail."""
        _run_aroom("workflow", "run", str(workflow_file), timeout=30)
        result = _run_aroom("workflow", "list")
        output = result.stdout
        assert result.returncode == 0
        # Run-centric columns should be present
        assert "Run" in output
        assert "Workflow" in output
        assert "State" in output
        # Step-level columns should NOT be primary headers
        assert "Step ID" not in output
        assert "Runner" not in output


# ---------------------------------------------------------------------------
# Watch subcommand (#888)
# ---------------------------------------------------------------------------


class TestWorkflowCLIWatch:
    """Integration tests for `aroom workflow watch`."""

    def test_watch_help(self) -> None:
        result = _run_aroom("workflow", "watch", "--help")
        assert result.returncode == 0
        assert "run_id" in result.stdout

    def test_watch_nonexistent_run(self) -> None:
        """Watch with a fake run ID shows error, doesn't crash."""
        result = _run_aroom("workflow", "watch", "nonexistent-fake-id", timeout=10)
        output = result.stdout + result.stderr
        assert "not found" in output.lower() or "Error" in output

    def test_watch_completed_run_exits(self, tmp_path: Path) -> None:
        """Watch exits immediately when the run is already in terminal status.

        Creates a real workflow run via `aroom workflow run`, then watches it.
        Since the run completes synchronously before watch starts, watch should
        see the terminal status and exit right away.
        """
        wf_path = tmp_path / "fast.yaml"
        wf_path.write_text(_TEST_WORKFLOW_YAML)

        # Run the workflow to completion
        run_result = _run_aroom("workflow", "run", str(wf_path), timeout=30)
        assert run_result.returncode == 0

        # Extract run_id from output — look for a UUID-like string after "run_id"
        import re

        run_output = run_result.stdout
        # The run output typically contains the run ID in the status display
        uuid_pattern = re.compile(r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}")
        match = uuid_pattern.search(run_output)
        if not match:
            pytest.skip("Could not extract run_id from workflow run output")

        run_id = match.group(0)

        # Watch the already-completed run — should exit immediately
        watch_result = _run_aroom("workflow", "watch", run_id, timeout=10)
        watch_output = watch_result.stdout + watch_result.stderr
        assert "completed" in watch_output.lower() or "failed" in watch_output.lower()


# ---------------------------------------------------------------------------
# Timeline output format (#1102)
# ---------------------------------------------------------------------------


def _run_workflow_and_get_id(workflow_yaml: str, tmp_path: Path) -> str:
    """Run a workflow and extract the run ID."""
    import re

    wf_path = tmp_path / "timeline_test.yaml"
    wf_path.write_text(workflow_yaml)
    run_result = _run_aroom("workflow", "run", str(wf_path), timeout=30)
    assert run_result.returncode == 0, f"workflow run failed: {run_result.stderr}"
    run_output = run_result.stdout
    uuid_pattern = re.compile(r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}")
    match = uuid_pattern.search(run_output)
    assert match, f"Could not extract run_id: {run_output}"
    return match.group(0)


class TestTimelineOutput:
    """Verify that watch/status/history use the timeline-style output (#1102)."""

    def test_status_shows_timeline_step_lines(self, tmp_path: Path) -> None:
        """Status output contains status icons and friendly durations."""
        run_id = _run_workflow_and_get_id(_TEST_WORKFLOW_YAML, tmp_path)
        result = _run_aroom("workflow", "status", run_id)
        output = result.stdout
        assert result.returncode == 0

        # Timeline step lines should show step names
        assert "greet" in output
        assert "validate" in output
        # Should show status icons (completed steps)
        # Rich renders Unicode so we check for the icon OR the step status text
        assert "completed" in output.lower() or "\u2705" in output or "success" in output.lower()
        # Should show the run header with workflow id
        assert "test_cli_pipeline" in output

    def test_status_shows_friendly_durations(self, tmp_path: Path) -> None:
        """Status uses friendly duration format, not raw ms."""
        run_id = _run_workflow_and_get_id(_TEST_WORKFLOW_YAML, tmp_path)
        result = _run_aroom("workflow", "status", run_id)
        output = result.stdout
        assert result.returncode == 0
        # Friendly durations contain 's' (e.g., '< 1s', '3s', '1m 5s')
        # The old format was raw 'ms' (e.g., '150ms')
        # At least one step should have completed with a friendly duration
        assert "s" in output  # every friendly duration ends in 's'

    def test_watch_completed_shows_timeline(self, tmp_path: Path) -> None:
        """Watch on a completed run shows timeline header and step lines."""
        run_id = _run_workflow_and_get_id(_TEST_WORKFLOW_YAML, tmp_path)
        result = _run_aroom("workflow", "watch", run_id, timeout=10)
        output = result.stdout + result.stderr
        assert "test_cli_pipeline" in output
        assert "greet" in output or "validate" in output
        # Should show terminal status
        assert "completed" in output.lower() or "failed" in output.lower()

    def test_history_preserves_detail_columns(self, tmp_path: Path) -> None:
        """History retains the detailed table with Runner, Result, Duration columns."""
        run_id = _run_workflow_and_get_id(_TEST_WORKFLOW_YAML, tmp_path)
        result = _run_aroom("workflow", "history", run_id)
        output = result.stdout
        assert result.returncode == 0
        # History should have the detail table columns (may be truncated by Rich)
        assert "Runner" in output
        assert "Result" in output
        assert "Durat" in output  # "Duration" or "Durat…" when truncated
        # Step names should be present
        assert "greet" in output
        assert "validate" in output
        # Should also show the timeline run header
        assert "test_cli_pipeline" in output

    def test_history_uses_friendly_durations(self, tmp_path: Path) -> None:
        """History table uses friendly duration format."""
        run_id = _run_workflow_and_get_id(_TEST_WORKFLOW_YAML, tmp_path)
        result = _run_aroom("workflow", "history", run_id)
        output = result.stdout
        assert result.returncode == 0
        # Old format was 'NNNms', new format is '< 1s' or 'Ns' etc.
        # The word 'ms' should NOT appear in duration column
        # (It may appear elsewhere like in column headers, but not as a step duration value)
        lines_with_greet = [ln for ln in output.splitlines() if "greet" in ln]
        assert len(lines_with_greet) > 0
        # The greet step's line should not have raw ms format
        for line in lines_with_greet:
            # Raw ms would look like '150ms' or '2000ms' — check it's not there
            import re

            assert not re.search(r"\d+ms", line), f"Raw ms duration found in history: {line}"


# A workflow with triggers for schedule tests (#969)
_TRIGGER_WORKFLOW_YAML = """\
kind: workflow
id: test_scheduled_pipeline
version: 0.1.0
inputs: {}
triggers:
  - type: schedule
    cron: "0 * * * *"
    target_ref: hourly-check
    missed_policy: skip
steps:
  - id: greet
    type: runner
    runner: shell
    command: "echo Hello from scheduled workflow"
    timeout: 10
"""


# ---------------------------------------------------------------------------
# Triggers and schedule subcommands (#969)
# ---------------------------------------------------------------------------


class TestWorkflowCLITriggers:
    """Integration tests for `aroom workflow triggers` and `aroom workflow schedule`."""

    def test_triggers_help(self) -> None:
        result = _run_aroom("workflow", "triggers", "--help")
        assert result.returncode == 0
        assert "list" in result.stdout
        assert "fire" in result.stdout

    def test_triggers_list_empty(self) -> None:
        result = _run_aroom("workflow", "triggers", "list")
        assert result.returncode == 0
        output = result.stdout + result.stderr
        assert "No workflow schedules" in output or "Workflow Schedules" in output

    def test_schedule_registers_triggers(self, tmp_path: Path) -> None:
        """Register triggers from a workflow YAML and verify they appear in triggers list."""
        # Place in ~/.anteroom/ so path confinement allows it
        anteroom_dir = tmp_path / ".anteroom"
        anteroom_dir.mkdir(exist_ok=True)
        wf_path = anteroom_dir / "scheduled.yaml"
        wf_path.write_text(_TRIGGER_WORKFLOW_YAML)

        # Register
        result = _run_aroom("workflow", "schedule", str(wf_path))
        assert result.returncode == 0
        output = result.stdout
        assert "Schedule created" in output

        # Verify it appears in the list
        list_result = _run_aroom("workflow", "triggers", "list")
        assert list_result.returncode == 0
        assert "0 * * * *" in list_result.stdout

    def test_schedule_help(self) -> None:
        result = _run_aroom("workflow", "schedule", "--help")
        assert result.returncode == 0
        assert "workflow_path" in result.stdout

    def test_triggers_enable_disable(self, tmp_path: Path) -> None:
        """Enable/disable a schedule via the CLI."""
        import re

        anteroom_dir = tmp_path / ".anteroom"
        anteroom_dir.mkdir(exist_ok=True)
        wf_path = anteroom_dir / "scheduled.yaml"
        wf_path.write_text(_TRIGGER_WORKFLOW_YAML)

        # Register first
        result = _run_aroom("workflow", "schedule", str(wf_path))
        assert result.returncode == 0

        # Extract full UUID from create output (stdout + stderr — Rich may write to either)
        all_output = result.stdout + result.stderr
        full_uuid = re.compile(r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}")
        full_match = full_uuid.search(all_output)
        if not full_match:
            pytest.skip("Could not extract full schedule_id from output")
        schedule_id = full_match.group(0)

        # Disable
        disable_result = _run_aroom("workflow", "triggers", "disable", schedule_id)
        assert disable_result.returncode == 0
        disable_output = disable_result.stdout + disable_result.stderr
        assert "disabled" in disable_output

        # Enable
        enable_result = _run_aroom("workflow", "triggers", "enable", schedule_id)
        assert enable_result.returncode == 0
        enable_output = enable_result.stdout + enable_result.stderr
        assert "enabled" in enable_output



"""Tests proving the real engine approval path: agent step → callback → pause → resume.

These tests use a mock AI service + mock tool executor to exercise the
actual WorkflowEngine approval callback, pause_signal, and resume path
without needing a real LLM endpoint.
"""

from __future__ import annotations

import tempfile
from pathlib import Path
from typing import Any
from unittest.mock import AsyncMock

import pytest

from anteroom.config import WorkflowConfig
from anteroom.db import init_db
from anteroom.services.ai_service import AIService
from anteroom.services.workflow_engine import WorkflowEngine, load_definition, register_gate_condition
from anteroom.services.workflow_runners import create_default_registry
from anteroom.services.workflow_storage import (
    get_pending_approval,
    list_workflow_steps,
    resolve_approval_request,
)

# Workflow with an agent runner step that will trigger tool approval
AGENT_WORKFLOW = """\
kind: workflow
id: test_agent_approval
version: 0.1.0
inputs: {}
steps:
  - id: write_code
    type: runner
    runner: cli_claude
    prompt: "Write a file"
    timeout: 30
"""


def _make_ai_service() -> AIService:
    service = AIService.__new__(AIService)
    service.config = type(
        "C",
        (),
        {
            "base_url": "http://localhost/v1",
            "api_key": "test",
            "model": "gpt-4",
            "request_timeout": 120,
            "verify_ssl": True,
            "max_output_tokens": 4096,
        },
    )()
    service._token_provider = None
    service.client = AsyncMock()
    return service


def _mock_ai_with_tool_call() -> AIService:
    """AI service that returns a write_file tool call."""
    service = _make_ai_service()
    call_count = 0

    async def fake_stream(messages: Any, **kwargs: Any):
        nonlocal call_count
        call_count += 1
        if call_count == 1:
            yield {
                "event": "tool_call",
                "data": {
                    "id": "tc1",
                    "function_name": "write_file",
                    "arguments": {"path": "/src/foo.py", "content": "# new file"},
                },
            }
            yield {"event": "done", "data": {}}
        else:
            yield {"event": "token", "data": {"content": "Done writing file."}}
            yield {"event": "done", "data": {}}

    service.stream_chat = fake_stream
    return service


def _make_approval_triggering_executor():
    """Tool executor that simulates a tool needing approval.

    When called with confirm_callback, it invokes the callback with
    needs_approval=True. If the callback returns False, it returns
    a denied result. If True, it returns success.
    """

    async def executor(name: str, args: dict[str, Any], **kwargs: Any) -> dict[str, Any]:
        confirm_callback = kwargs.get("confirm_callback")
        if confirm_callback is not None:
            # Create a mock SafetyVerdict
            verdict = type(
                "SafetyVerdict",
                (),
                {
                    "tool_name": name,
                    "needs_approval": True,
                    "is_hard_blocked": False,
                    "reason": "Write operation requires approval",
                    "details": args,
                    "tier": "WRITE",
                },
            )()
            approved = await confirm_callback(verdict)
            if not approved:
                return {"error": "Operation denied by user", "_approval_decision": "denied"}
        return {"result": f"{name} executed", "_approval_decision": "auto"}

    return executor


@pytest.fixture()
def db():
    with tempfile.TemporaryDirectory() as td:
        conn = init_db(Path(td) / "test.db")
        yield conn
        conn.close()


@pytest.fixture(autouse=True)
def _register_gates():
    async def always_pass(run: Any, step: Any, inputs: Any) -> bool:
        return True

    register_gate_condition("always_pass", always_pass)
    yield


class TestRealApprovalPath:
    """Prove the engine's actual approval callback fires and pauses correctly."""

    @pytest.mark.asyncio
    async def test_agent_step_pauses_on_approval(self, db: Any) -> None:
        """Agent runner step triggers approval → run enters waiting_for_approval."""
        ai = _mock_ai_with_tool_call()
        executor = _make_approval_triggering_executor()

        engine = WorkflowEngine(
            db,
            WorkflowConfig(),
            create_default_registry(),
            ai_service=ai,
            tool_executor=executor,
            tools_openai=[{"type": "function", "function": {"name": "write_file"}}],
        )

        defn = load_definition(AGENT_WORKFLOW)
        run = await engine.start_run(defn, target_kind="task", target_ref="t1")

        # The run should be paused for approval, not completed
        assert run["status"] == "waiting_for_approval", (
            f"Expected waiting_for_approval, got {run['status']} (stop_reason: {run.get('stop_reason')})"
        )

        # A pending approval request should exist
        pending = get_pending_approval(db, run["id"])
        assert pending is not None
        assert pending["tool_name"] == "write_file"
        assert pending["risk_tier"] == "WRITE"
        assert pending["status"] == "pending"

    @pytest.mark.asyncio
    async def test_approval_request_id_stored_on_step(self, db: Any) -> None:
        """The approval request ID is stored on the step record."""
        ai = _mock_ai_with_tool_call()
        executor = _make_approval_triggering_executor()

        engine = WorkflowEngine(
            db,
            WorkflowConfig(),
            create_default_registry(),
            ai_service=ai,
            tool_executor=executor,
            tools_openai=[{"type": "function", "function": {"name": "write_file"}}],
        )

        defn = load_definition(AGENT_WORKFLOW)
        run = await engine.start_run(defn, target_kind="task", target_ref="t2")

        steps = list_workflow_steps(db, run["id"])
        assert len(steps) >= 1
        # The step that triggered approval should have approval_request_id
        approval_step = next((s for s in steps if s.get("approval_request_id")), None)
        assert approval_step is not None, f"No step has approval_request_id. Steps: {steps}"

    @pytest.mark.asyncio
    async def test_approve_then_resume_completes(self, db: Any) -> None:
        """Approve the request, resume the run → run completes."""
        ai = _mock_ai_with_tool_call()
        executor = _make_approval_triggering_executor()

        engine = WorkflowEngine(
            db,
            WorkflowConfig(),
            create_default_registry(),
            ai_service=ai,
            tool_executor=executor,
            tools_openai=[{"type": "function", "function": {"name": "write_file"}}],
        )

        defn = load_definition(AGENT_WORKFLOW)
        run = await engine.start_run(defn, target_kind="task", target_ref="t3")
        assert run["status"] == "waiting_for_approval"

        # Approve the request
        pending = get_pending_approval(db, run["id"])
        resolve_approval_request(db, pending["id"], status="approved", resolved_by="operator")

        # Resume — the pre-approved callback should let the tool through
        result = await engine.resume_run(run["id"], defn)
        assert result["status"] == "completed", (
            f"Expected completed, got {result['status']} (stop_reason: {result.get('stop_reason')})"
        )

    @pytest.mark.asyncio
    async def test_deny_then_resume_cancels(self, db: Any) -> None:
        """Deny the request, resume → run is cancelled."""
        ai = _mock_ai_with_tool_call()
        executor = _make_approval_triggering_executor()

        engine = WorkflowEngine(
            db,
            WorkflowConfig(),
            create_default_registry(),
            ai_service=ai,
            tool_executor=executor,
            tools_openai=[{"type": "function", "function": {"name": "write_file"}}],
        )

        defn = load_definition(AGENT_WORKFLOW)
        run = await engine.start_run(defn, target_kind="task", target_ref="t4")
        assert run["status"] == "waiting_for_approval"

        pending = get_pending_approval(db, run["id"])
        resolve_approval_request(db, pending["id"], status="denied", resolved_by="operator")

        result = await engine.resume_run(run["id"], defn)
        assert result["status"] == "cancelled"
        assert "denied" in (result.get("stop_reason") or "")



"""Tests for workflow step templating — overridable params (#958)."""

from __future__ import annotations

import tempfile
from pathlib import Path
from typing import Any
from unittest.mock import AsyncMock, patch

import pytest

from anteroom.db import init_db
from anteroom.services.workflow_engine import (
    WorkflowEngine,
    _resolve_params,
    load_definition,
    resolve_template,
)
from anteroom.services.workflow_runners import create_default_registry
from anteroom.services.workflow_storage import (
    create_workflow_run,
    get_workflow_run,
)

# ---------------------------------------------------------------------------
# YAML fixtures
# ---------------------------------------------------------------------------

YAML_WITH_PARAMS = """\
kind: workflow
id: test_params
version: "1.0"
params:
  lint_command:
    default: "ruff check src/"
    description: "Lint command"
  test_command:
    default: "pytest tests/"
    description: "Test command"
inputs:
  issue_number:
    required: true
    type: string
steps:
  - id: checks
    type: runner
    runner: shell
    command: "{lint_command} && {test_command}"
  - id: report
    type: runner
    runner: shell
    command: "echo issue {issue_number}"
"""

YAML_WITH_REQUIRED_PARAM = """\
kind: workflow
id: test_required_param
version: "1.0"
params:
  deploy_target:
    description: "Deploy target — no default, must be provided"
steps:
  - id: deploy
    type: runner
    runner: shell
    command: "deploy.sh {deploy_target}"
"""

YAML_PARAM_INPUT_OVERLAP = """\
kind: workflow
id: test_overlap
version: "1.0"
params:
  issue_number:
    default: "99"
inputs:
  issue_number:
    required: true
steps:
  - id: check
    type: runner
    runner: shell
    command: "echo {issue_number}"
"""

YAML_NO_PARAMS = """\
kind: workflow
id: test_no_params
version: "1.0"
steps:
  - id: hello
    type: runner
    runner: shell
    command: "echo hello"
"""

YAML_EMPTY_PARAMS = """\
kind: workflow
id: test_empty_params
version: "1.0"
params: {}
steps:
  - id: hello
    type: runner
    runner: shell
    command: "echo hello"
"""


# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------


@pytest.fixture()
def db():
    with tempfile.TemporaryDirectory() as td:
        conn = init_db(Path(td) / "test.db")
        yield conn
        conn.close()


@pytest.fixture()
def engine(db: Any) -> WorkflowEngine:
    from anteroom.config import WorkflowConfig

    config = WorkflowConfig()
    registry = create_default_registry()
    return WorkflowEngine(db, config, registry)


# ---------------------------------------------------------------------------
# Parsing tests
# ---------------------------------------------------------------------------


class TestParamParsing:
    def test_load_definition_with_params(self) -> None:
        defn = load_definition(YAML_WITH_PARAMS)
        assert "lint_command" in defn.params
        assert defn.params["lint_command"]["default"] == "ruff check src/"
        assert defn.params["lint_command"]["description"] == "Lint command"
        assert "test_command" in defn.params

    def test_load_definition_no_params(self) -> None:
        defn = load_definition(YAML_NO_PARAMS)
        assert defn.params == {}

    def test_load_definition_empty_params(self) -> None:
        defn = load_definition(YAML_EMPTY_PARAMS)
        assert defn.params == {}

    def test_load_definition_required_param(self) -> None:
        defn = load_definition(YAML_WITH_REQUIRED_PARAM)
        assert "deploy_target" in defn.params
        assert "default" not in defn.params["deploy_target"]


class TestParamInputOverlap:
    def test_overlap_rejected(self) -> None:
        with pytest.raises(ValueError, match="overlap with input keys"):
            load_definition(YAML_PARAM_INPUT_OVERLAP)


# ---------------------------------------------------------------------------
# Resolution tests
# ---------------------------------------------------------------------------


class TestResolveParams:
    def test_defaults_used_when_no_overrides(self) -> None:
        defn = load_definition(YAML_WITH_PARAMS)
        resolved = _resolve_params(defn)
        assert resolved == {
            "lint_command": "ruff check src/",
            "test_command": "pytest tests/",
        }

    def test_overrides_win_over_defaults(self) -> None:
        defn = load_definition(YAML_WITH_PARAMS)
        resolved = _resolve_params(defn, {"lint_command": "eslint src/"})
        assert resolved["lint_command"] == "eslint src/"
        assert resolved["test_command"] == "pytest tests/"

    def test_required_param_missing_raises(self) -> None:
        defn = load_definition(YAML_WITH_REQUIRED_PARAM)
        with pytest.raises(ValueError, match="Required param 'deploy_target'"):
            _resolve_params(defn)

    def test_required_param_supplied(self) -> None:
        defn = load_definition(YAML_WITH_REQUIRED_PARAM)
        resolved = _resolve_params(defn, {"deploy_target": "staging"})
        assert resolved["deploy_target"] == "staging"

    def test_empty_params_returns_empty_dict(self) -> None:
        defn = load_definition(YAML_EMPTY_PARAMS)
        resolved = _resolve_params(defn)
        assert resolved == {}

    def test_override_value_coerced_to_string(self) -> None:
        defn = load_definition(YAML_WITH_PARAMS)
        resolved = _resolve_params(defn, {"lint_command": 42})  # type: ignore[dict-item]
        assert resolved["lint_command"] == "42"


# ---------------------------------------------------------------------------
# Template expansion tests
# ---------------------------------------------------------------------------


class TestTemplateExpansionWithParams:
    def test_params_and_inputs_merged_in_template(self) -> None:
        defn = load_definition(YAML_WITH_PARAMS)
        resolved = _resolve_params(defn)
        template_vars = {**resolved, "issue_number": "123"}
        result = resolve_template("{lint_command} && {test_command} #{issue_number}", template_vars)
        assert result == "ruff check src/ && pytest tests/ #123"

    def test_shell_quoting_applied_to_params(self) -> None:
        defn = load_definition(YAML_WITH_PARAMS)
        resolved = _resolve_params(defn, {"lint_command": "ruff check 'src dir'"})
        template_vars = {**resolved, "issue_number": "1"}
        result = resolve_template("{lint_command}", template_vars, shell_quote=True)
        assert "'" in result  # shlex.quote wraps the value


# ---------------------------------------------------------------------------
# Persistence tests
# ---------------------------------------------------------------------------


class TestParamPersistence:
    def test_params_stored_in_run_record(self, db: Any) -> None:
        params = {"lint_command": "eslint src/", "test_command": "jest"}
        run = create_workflow_run(
            db,
            workflow_id="test",
            workflow_version="1.0",
            target_kind="workflow",
            target_ref="test",
            params=params,
        )
        assert run["params"] == params

    def test_params_round_trip_via_get(self, db: Any) -> None:
        params = {"lint_command": "eslint src/"}
        run = create_workflow_run(
            db,
            workflow_id="test",
            workflow_version="1.0",
            target_kind="workflow",
            target_ref="test",
            params=params,
        )
        fetched = get_workflow_run(db, run["id"])
        assert fetched is not None
        assert fetched["params"] == params

    def test_null_params_for_no_params(self, db: Any) -> None:
        run = create_workflow_run(
            db,
            workflow_id="test",
            workflow_version="1.0",
            target_kind="workflow",
            target_ref="test",
        )
        assert run["params"] is None

    def test_params_deserialized_from_json(self, db: Any) -> None:
        params = {"key": "value with spaces"}
        run = create_workflow_run(
            db,
            workflow_id="test",
            workflow_version="1.0",
            target_kind="workflow",
            target_ref="test",
            params=params,
        )
        fetched = get_workflow_run(db, run["id"])
        assert fetched is not None
        assert fetched["params"]["key"] == "value with spaces"


# ---------------------------------------------------------------------------
# Engine integration tests
# ---------------------------------------------------------------------------


@pytest.fixture(autouse=True)
def _register_test_gates():
    from anteroom.services.workflow_engine import register_gate_condition

    async def always_pass(run: Any, step: Any, inputs: Any) -> bool:
        return True

    register_gate_condition("always_pass", always_pass)
    yield


class TestEngineParamIntegration:
    @pytest.mark.asyncio
    async def test_start_run_resolves_params(self, engine: WorkflowEngine, db: Any) -> None:
        defn = load_definition(YAML_WITH_PARAMS)

        with patch(
            "anteroom.services.workflow_runners.execute_opaque_runner",
            new_callable=AsyncMock,
        ) as mock_runner:
            from anteroom.services.workflow_runners import RunnerResult

            mock_runner.return_value = RunnerResult(status="success", summary="ok")

            run = await engine.start_run(
                defn,
                target_kind="workflow",
                target_ref="test",
                inputs={"issue_number": "42"},
                param_overrides={"lint_command": "eslint ."},
            )

        assert run["status"] == "completed"
        # Verify the runner was called with the resolved param in the command
        calls = mock_runner.call_args_list
        assert len(calls) == 2
        first_call_kwargs = calls[0][1]
        assert "eslint ." in first_call_kwargs.get("command", "")

    @pytest.mark.asyncio
    async def test_start_run_persists_params(self, engine: WorkflowEngine, db: Any) -> None:
        defn = load_definition(YAML_WITH_PARAMS)

        with patch(
            "anteroom.services.workflow_runners.execute_opaque_runner",
            new_callable=AsyncMock,
        ) as mock_runner:
            from anteroom.services.workflow_runners import RunnerResult

            mock_runner.return_value = RunnerResult(status="success", summary="ok")

            run = await engine.start_run(
                defn,
                target_kind="workflow",
                target_ref="test_persist",
                inputs={"issue_number": "99"},
                param_overrides={"test_command": "jest"},
            )

        fetched = get_workflow_run(db, run["id"])
        assert fetched is not None
        assert fetched["params"]["test_command"] == "jest"
        assert fetched["params"]["lint_command"] == "ruff check src/"

    @pytest.mark.asyncio
    async def test_start_run_no_params(self, engine: WorkflowEngine, db: Any) -> None:
        defn = load_definition(YAML_NO_PARAMS)

        with patch(
            "anteroom.services.workflow_runners.execute_opaque_runner",
            new_callable=AsyncMock,
        ) as mock_runner:
            from anteroom.services.workflow_runners import RunnerResult

            mock_runner.return_value = RunnerResult(status="success", summary="ok")

            run = await engine.start_run(
                defn,
                target_kind="workflow",
                target_ref="test_no_params",
            )

        assert run["status"] == "completed"

    @pytest.mark.asyncio
    async def test_start_run_required_param_missing_raises(self, engine: WorkflowEngine) -> None:
        defn = load_definition(YAML_WITH_REQUIRED_PARAM)

        with pytest.raises(ValueError, match="Required param"):
            await engine.start_run(
                defn,
                target_kind="workflow",
                target_ref="test_required",
            )

    @pytest.mark.asyncio
    async def test_enqueue_run_resolves_params(self, engine: WorkflowEngine, db: Any) -> None:
        defn = load_definition(YAML_WITH_PARAMS)

        run = await engine.enqueue_run(
            defn,
            target_kind="workflow",
            target_ref="test_enqueue",
            inputs={"issue_number": "7"},
            param_overrides={"lint_command": "mypy src/"},
        )

        assert run["status"] == "pending"
        fetched = get_workflow_run(db, run["id"])
        assert fetched is not None
        assert fetched["params"]["lint_command"] == "mypy src/"
        assert fetched["params"]["test_command"] == "pytest tests/"
